/*********************************************************************
 *                
 * Filename:      driver.c
 * Version:       
 * Description:   
 * Status:        Experimental.
 * Author:        Dag Brattli <dagb@cs.uit.no>
 * Created at:    Wed Sep  2 20:22:08 1998
 * Modified at:   Tue Oct 20 10:41:16 1998
 * Modified by:   Dag Brattli <dagb@cs.uit.no>
 * 
 *     Copyright (c) 1998 Dag Brattli, All Rights Reserved.
 *      
 *     This program is free software; you can redistribute it and/or 
 *     modify it under the terms of the GNU General Public License as 
 *     published by the Free Software Foundation; either version 2 of 
 *     the License, or (at your option) any later version.
 *  
 *     Neither Dag Brattli nor University of Troms admit liability nor
 *     provide warranty for any of this software. This material is 
 *     provided "AS-IS" and at no charge.
 *     
 ********************************************************************/

#include <linux/proc_fs.h>

#include "driver.h"
#include "timer.h"
#include "wrapper.h"

HASHBIN *driver = NULL;

#ifdef USE_PROC
#ifdef CONFIG_PROC_FS
int driver_proc_read( char *buf, char **start, off_t offset, int len, 
		      int unused);

#endif /* CONFIG_PROC_FS */
#endif /* USE_PROC */

int driver_init( void)
{
	DEBUG( 4, __FUNCTION__ "()\n");

	/* Allocate master array */
	driver = hashbin_new( HB_LOCAL);
	if ( driver == NULL) {
		printk( KERN_WARNING "IrDA: Can't allocate driver hashbin!\n");
		return -ENOMEM;
	}
	return 0;
}

void driver_cleanup(void)
{
	DEBUG( 4, __FUNCTION__ "()\n");

	ASSERT( driver != NULL, return;);

	hashbin_delete( driver, (FREE_FUNC) driver_close);
}

/*
 * Function driver_open (self)
 *
 *    
 *
 */
int driver_open( struct driver_cb *self) 
{        
	DEBUG( 4, __FUNCTION__ "()\n");     

	self->rx.buff = ( __u8 *) kmalloc( BUF_MAX_LEN, GFP_KERNEL);
	self->tx.buff = ( __u8 *) kmalloc( BUF_MAX_LEN, GFP_KERNEL);

	if ( self->rx.buff == NULL || self->tx.buff == NULL)   {
		DEBUG( 0, "IrDA Self: no space for buffers!\n");
		driver_close( self);
		return -ENOMEM;
	}

	memset( self->rx.buff, 0, BUF_MAX_LEN);
	memset( self->tx.buff, 0, BUF_MAX_LEN);
	
	self->magic = DRIVER_MAGIC;

	self->rx.buffsize = BUF_MAX_LEN;
	self->tx.buffsize = BUF_MAX_LEN;

	skb_queue_head_init( &self->rx.list);
	skb_queue_head_init( &self->tx.list);

	self->rx.in_frame = FALSE;
	self->rx.state = OUTSIDE_FRAME;

	irlap_open( self);

	hashbin_insert( driver, (QUEUE *) self, (int) self, NULL);

	printk( "IrDA driver %s registered.\n", self->name);
        
	return 0;
}

/*
 * Function driver_close (self)
 *
 *    Close this instance of the driver, just deallocate buffers
 *
 */
void __driver_close( struct driver_cb *self)
{
	struct sk_buff *skb;

	DEBUG( 4, __FUNCTION__ "()\n");

	ASSERT( self != NULL, return;);
	ASSERT( self->magic == DRIVER_MAGIC, return;);

	if ( self->tx.buff) {
		kfree( self->tx.buff);
	}

	if ( self->rx.buff) {
		kfree( self->rx.buff); 
	}

	while (( skb = skb_dequeue( &self->rx.list)) != NULL) {
		DEV_KFREE_SKB( skb, FREE_READ);  
	}

	while (( skb = skb_dequeue( &self->tx.list)) != NULL) {
		DEV_KFREE_SKB( skb, FREE_WRITE); 
	}

}

/*
 * Function driver_close (self)
 *
 *    
 *
 */
void driver_close( struct driver_cb *self)
{
	DEBUG( 4, __FUNCTION__ "()\n");

	ASSERT( self != NULL, return;);
	ASSERT( self->magic == DRIVER_MAGIC, return;);

	/* Stop IrLAP */
	irlap_close( self->irlap);

	hashbin_remove( driver, (int) self, NULL);

	__driver_close( self);
}

/*
 * Function driver_queue_xmit (self, skb)
 *
 *    
 *
 */
inline int driver_queue_xmit( struct driver_cb *self, struct sk_buff *skb)
{
	/*
	 *  Check if driver exists
	 */
	if ( self == NULL) {
		DEBUG( 0, "driver_queue_xmit(), driver is missing!\n");
		return -1;
	}

	ASSERT( self->is_tbusy != NULL, return -1;);

	/* 
	 *  If driver is busy transmitting or media is busy then queue
	 *  packet for transmission later. We don't check if the media
	 *  is busy, since queueing the frame will only make it worse since
	 *  that frame will have to wait until the next transmission, and it
	 *  will then be transmitted as the last frame, which confuses the
	 *  peer state machine a lot! Check for media busy should be done
	 *  in the irlap state machine. DB.
	 */
	/* if ( self->is_tbusy( self) || driver_is_media_busy( self)) { */
	
	if ( self->is_tbusy( self)) {
		DEBUG( 4, "driver: queuing skb for later transmission!\n");
		skb_queue_tail( &self->tx.list, skb);
		
		return 0;
	}
	
	DEBUG( 6, "driver_queue_xmit: sending skb on!\n");
	self->hard_xmit( self, skb);
	
	return 0;
}

/*
 * Function driver_set_media_busy (self, status)
 *
 *    Called when we have detected that another station is transmiting
 *    in contention mode.
 */
void driver_set_media_busy( struct driver_cb *self, int status) 
{
	DEBUG( 4, "driver_set_media_busy: (%s)\n", status ? "TRUE" : "FALSE");

	ASSERT( self != NULL, return;);
	ASSERT( self->magic == DRIVER_MAGIC, return;);

	if ( status) {
		self->media_busy = TRUE;
		driver_start_mbusy_timer( self);
		DEBUG( 4, "Media busy!\n");
	} else {
		self->media_busy = FALSE;
		del_timer( &self->media_busy_timer);
	}
}

inline void driver_change_speed( struct driver_cb *self, int speed)
{
	ASSERT( self != NULL, return;);
	ASSERT( self->magic == DRIVER_MAGIC, return;);

	if ( self->change_speed)
		self->change_speed( self, speed);
}

inline int driver_is_media_busy( struct driver_cb *self)
{
	ASSERT( self != NULL, return FALSE;);
	ASSERT( self->magic == DRIVER_MAGIC, return FALSE;);
	
	return self->media_busy;
}

inline int driver_is_receiving( struct driver_cb *self)
{
	ASSERT( self != NULL, return FALSE;);
	ASSERT( self->magic == DRIVER_MAGIC, return FALSE;);

	DEBUG( 4, __FUNCTION__ " (%s)\n",  
	       ( self->rx.state != OUTSIDE_FRAME) ? "TRUE" : "FALSE");

	return ( self->rx.state != OUTSIDE_FRAME);
}

inline struct qos_info *driver_get_qos( struct driver_cb *self)
{
	ASSERT( self != NULL, return NULL;);
	ASSERT( self->magic == DRIVER_MAGIC, return NULL;);

	return &self->qos;
}

#ifdef USE_PROC
#ifdef CONFIG_PROC_FS
/*
 * Function irlap_proc_read (buf, start, offset, len, unused)
 *
 *    Give some info to the /proc file system
 *
 */
int driver_proc_read( char *buf, char **start, off_t offset, int len, 
		      int unused)
{
	struct driver_cb *self;
	unsigned long flags;
     
	save_flags(flags);
	cli();

	len = 0;

	self = (struct driver_cb *) hashbin_get_first( driver);
	while ( self != NULL) {
		len += sprintf( buf+len, "Driver name: %s\n", self->name);
		len += sprintf( buf+len, "  bps\tmaxtt\tdsize\twinsize\taddbofs\tmintt\tldisc\n");
		
		len += sprintf( buf+len, "  %d\t", 
				self->qos.baud_rate.value);
		len += sprintf( buf+len, "%d\t", 
				self->qos.max_turn_time.value);
		len += sprintf( buf+len, "%d\t",
				self->qos.data_size.value);
		len += sprintf( buf+len, "%d\t",
				self->qos.window_size.value);
		len += sprintf( buf+len, "%d\t",
				self->qos.additional_bofs.value);
		len += sprintf( buf+len, "%d\t", 
				self->qos.min_turn_time.value);
		len += sprintf( buf+len, "%d", 
				self->qos.link_disc_time.value);
		len += sprintf( buf+len, "\n");
	       
		self = (struct driver_cb *) hashbin_get_next( driver);
	}
	restore_flags(flags);

	return len;
}

#endif /* CONFIG_PROC_FS */
#endif /* USE_PROC */

