/*********************************************************************
 *		  
 * Filename:	  irport.c
 * Version:	  0.1
 * Description:  Some kind of IR thinger.
 * Status:	  Experimental.
 * Author:	  Ryan Shillington
 ********************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <linux/errno.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/serial_reg.h>
#include <linux/delay.h>
#include <linux/tqueue.h>

#include <linux/fs.h>

#include "driver.h"
#include "irport.h"
#include "crc.h"
#include "wrapper.h"
#include "ircompat.h"
#include "w83977af_ir.h"
#include "irlap_frame.h"
#include "irlap.h"

#define IO_EXTENT	8

/* How long before we timeout */
#define MEDIABUSY_TIMEOUT 1000

static void irport_interrupt(int irq, void *dev_id, struct pt_regs *regs);

int irport_is_receiving( struct fir_driver_cb *self);
void irport_start_media_timer( struct driver_cb *driver);
int irport_is_tbusy(struct driver_cb *driver);
void irport_media_busy_expired( unsigned long data);
void irport_todo_expired( unsigned long data);
void irport_start_todo_timer( struct driver_cb *driver, int timeout);

static char *driver_name = "irport_w83977af";
struct fir_driver_cb *irport_drv;
struct tq_struct irport_task;

/*
 * Initialize IO port 
 */
int irport_init( void) 
{
	struct fir_driver_cb *fir_driver;
	struct driver_cb *driver;
  int result;
        
	DEBUG(4, "irport_init - START\n");

  /* 1 - Initialize the driver */
	fir_driver = kmalloc(sizeof(struct fir_driver_cb), GFP_KERNEL);
	memset(fir_driver, 0, sizeof(struct fir_driver_cb));

  /* 2 - Keep track of the driver (put it in a global variable) */
	irport_drv = fir_driver;
	driver = &fir_driver->drv;

  /* 3 - Set the specific information about the hardware */
	fir_driver->baud = 9600;
  driver->name = driver_name;
  driver->magic = DRIVER_MAGIC;
	fir_driver->SCE.iobase = 0x400;
	fir_driver->SCE.irq = 22;
	fir_driver->SCE.ioextent = 8;
	fir_driver->SCE.dma = 0;
	fir_driver->SCE.fifo_size = 16;
	driver->hard_xmit = irport_xmit;
	driver->change_speed = irport_setspeed;
	fir_driver->irq_handler = irport_interrupt;
  fir_driver->tbusy = 0;
	driver->is_tbusy = irport_is_tbusy;
  driver->rx.state = OUTSIDE_FRAME;
  driver->self = fir_driver;
#if 1
  driver->driver_flags |= SIR_MODE | IO_PIO;
#else
  driver->driver_flags |= SIR_MODE | MIR_MODE | FIR_MODE | IO_DMA;
#endif
  driver->name = driver_name;

#if 1
  driver->qos.baud_rate.bits = IR_9600|IR_19200|IR_38400|IR_57600|IR_115200;
  driver->qos.window_size.bits     = 0x7f; /* 7 max */
  driver->qos.min_turn_time.bits   = 0x0f;
  driver->qos.max_turn_time.bits   = 0x01; /* 500ms for now */
  driver->qos.data_size.bits       = 0x3f; /* 2048 max */
  driver->qos.link_disc_time.bits  = 0xff;
  driver->qos.additional_bofs.bits = 0xff; /* 0 min BOFs */
  irda_qos_bits_to_value( &driver->qos);
#else
  driver->qos.baud_rate = IR_9600|IR_19200|IR_38400|IR_57600|IR_115200
    |IR_576000|IR_1152000|(IR_4000000 << 8);
#endif

  /* 4 - Allocate and initialize memory for Transmit and Receive */
  result = driver_open(driver);
  if (result !=0)
  {
    DEBUG(0, "Could not complete driver_open. Status received =%d\n", result);
    return -1;
  }

  /* 5 - Detect to make sure that the device exists */
	if (FIR_detect(fir_driver) < 0) {
		DEBUG(0, "Error, device not found!\n");
		return -1;
	}
  
  /* 6 - Lock the port that we need */
  if (check_region(fir_driver->SCE.iobase, fir_driver->SCE.ioextent) < 0) 
  { 
    DEBUG(0, "IrVDD: can't get SCE iobase of 0x%03x\n", fir_driver->SCE.iobase);
    irport_cleanup(fir_driver); 
    return -ENODEV;
  } else {
    request_region( fir_driver->SCE.iobase,
                    fir_driver->SCE.ioextent,
                    fir_driver->drv.name);
  }

  /* 7 - Register the IRQ handler */
  if(fir_driver->SCE.irq > 0)
  {
    request_irq(fir_driver->SCE.irq, fir_driver->irq_handler, SA_INTERRUPT,
      fir_driver->drv.name, (void *) fir_driver);
  }

  /* 8 - The device is initially NOT busy */
	irport_set_media_busy(driver, FALSE); 

  /* 9 - This starts up the SIR stuff?? */
	irport_start(fir_driver);

  /* 10 - Set the speed of the IR device */
	irport_setspeed( driver, 9600 ); 

  /* 11 - Print out a message to confirm that this function is finished */
	DEBUG(4, "irport_init - FINISHED\n");

	return 0;
}

/*
 * Function irport_cleanup ()
 *
 *    irport_cleanup(): Removes everything!
 *
 */
void irport_cleanup(struct fir_driver_cb *driver) 
{
	int SCE_iobase;

	DEBUG(4, "irport_cleanup - START\n");

  /* 1 - Initialize important variables */
	SCE_iobase = driver->SCE.iobase;

  /* 2 - Shutdown the device */
	FIR_shutdown(driver);

	/* 3 - Reset uart */
	outb_p( 0, SCE_iobase+UART_MCR);

	/* 4 - Turn off interrupts */
  DEBUG(4, "Turning off interupts\n");
	outb_p( 0, SCE_iobase+UART_IER); 

  /* 5 - Unregister the interrupts */
  DEBUG(4, "Releasing IRQ %d\n", driver->SCE.irq);
  if(driver->SCE.irq >0)
    free_irq(driver->SCE.irq, driver);

  /* 6 - Release the PORT that this driver is using */
  DEBUG(4, "Releasing Region %03x\n", driver->SCE.iobase);
  release_region( driver->SCE.iobase, driver->SCE.ioextent);

	/* 7 - Stop media busy timer */
	del_timer( &driver->drv.media_busy_timer);
	del_timer( &driver->drv.todo_timer);

  /* 8 - Empty tx_list and rx_list! */
  driver_cleanup();
  
  /* 9 - Delete the driver out of memory (This must be last) */
  if (driver)
        kfree_s(driver, sizeof(struct fir_driver_cb));

}

/*
 * Function irport_start (void)
 *
 *    Start IO port 
 *
 */
int irport_start(struct fir_driver_cb *driver) 
{
	int iobase = driver->SCE.iobase;

  /*
   * Reset the UART.
   */

  outb_p( 0x00, iobase+UART_MCR);
  outb_p((UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT), iobase+UART_FCR);
  (void)inb( iobase+UART_RX);

	/* Initialize UART */
	outb_p( UART_LCR_WLEN8, iobase+UART_LCR);  /* Reset DLAB */
	outb_p(( UART_MCR_DTR | UART_MCR_RTS | UART_MCR_OUT2), iobase+UART_MCR);

	/* Turn on interrups */
	outb_p(( UART_IER_THRI |UART_IER_RLSI | UART_IER_RDI), iobase+UART_IER); 
	return 0;
}

/*
 * Function irport_setspeed (speed)
 *
 *    Set speed of port to specified baudrate
 *
 */
void irport_setspeed( struct driver_cb *driver, int speed) 
{
  struct fir_driver_cb *fir_driver = (struct fir_driver_cb *) driver->self;
	int iobase = fir_driver->SCE.iobase;
	int fcr;    /* FIFO control reg */
	int lcr;    /* Line control reg */
	int divisor;

  DEBUG( 4, __FUNCTION__ "()\n");
  ASSERT( driver != NULL, return;);
  ASSERT( driver->magic == DRIVER_MAGIC, return;);

	if (speed > 115200) {
		FIR_setspeed(fir_driver, speed);
	} else {
     /*
      *  If we are transmitting, or if the driver has more bytes to send,
      *  then we must wait until the frame is sent before changing the
      *  speed of the serial port. The test is not quite accurate since the
      *  fifo may still contain bytes to send. 50 msecs should be enough
      *  to make it behave correctly, see calculation below:
      *  
      *  23 bytes SNRM * 2 for stuffing + 11 bofs + 2 CRC = 59 bytes
      *  59 bytes / 9600 bps = 49 ms.
      * 
      *  Some stations may take 200 ms or more before they can operate on
      *  a new baudrate. (note 5 IrLAP p. 60)
      */
    if ((test_bit(0, &fir_driver->tbusy)) 
        || (skb_queue_len( &driver->tx.list)))
    {
      DEBUG(3, "Waiting 50ms to set baud rate...\n");
      fir_driver->new_speed = speed;
      irport_start_todo_timer( driver, 5);
      return;
    }
        
		divisor = SPEED_MAX / speed;

		fcr = UART_FCR_ENABLE_FIFO | UART_FCR_TRIGGER_14;

		/* IrDA ports use 8N1 */
		lcr = UART_LCR_WLEN8;

	DEBUG(0, "Starting %s: Setting speed to: %d\n", driver->name, speed);
    irda_lock(&fir_driver->sbusy);

		outb_p( UART_LCR_DLAB | lcr, iobase+UART_LCR); /* Set DLAB */
		outb_p( divisor & 0xff,      iobase+UART_DLL); /* Set speed	*/
		outb_p( divisor >> 8,	     iobase+UART_DLM);
		outb_p( lcr,		     iobase+UART_LCR); /* Set 8N1	*/
		outb_p( fcr,		     iobase+UART_FCR); /* Enable FIFO's */
    irda_unlock(&fir_driver->sbusy);
	}

	DEBUG(0, "%s: Setting speed to: %d\n", driver->name, speed);

	fir_driver->baud = speed;
  fir_driver->new_speed = 0;
}

/*
 * Function irport_interrupt (irq, dev_id, regs)
 *
 *    
 */
static void irport_interrupt(int irq, void *dev_id, struct pt_regs *regs) 
{
	struct fir_driver_cb *driver = (struct fir_driver_cb *) dev_id;

	int iobase, status;
	int iir;

	DEBUG(5, "irport_interrupt: irq %d\n", irq);

	if (!driver) {
		printk(KERN_WARNING "irport_interrupt: irq %d for unknown device.\n", irq);
		return;
	}
	
	if (driver->baud > 115200) {
		FIR_interrupt(irq, dev_id, regs);
	} else {
		iobase = driver->SCE.iobase;
		iir = inb( iobase+UART_IIR);
		do {
			status = inb( iobase+UART_LSR);

			if ( status & UART_LSR_DR) {
				/* Receive interrupt */
        DEBUG(5, "Received Data\n");
				irport_receive(driver);
			}
			if ( status & UART_LSR_THRE) {
				/* Transmitter ready for data */
        DEBUG(5, "Ready to Transmit Data\n");
				irport_write(&driver->drv);
			}
		} while ( !(inb( iobase+UART_IIR) & UART_IIR_NO_INT));
	}

	return;

}

int irport_write(struct driver_cb *driver)
{
  struct fir_driver_cb *fir_driver = (struct fir_driver_cb *) driver->self;
	int iobase = fir_driver->SCE.iobase;
	int fifo_size, cnt, count=0;
	struct sk_buff *skb;

	if (!(inb_p( iobase+UART_LSR) & UART_LSR_THRE)) {
		DEBUG(0, "irport_write failed, fifo not empty! Bytes left = %d, Packets = %d",
        driver->tx.count-driver->tx.ptr, skb_queue_len( &driver->tx.list));
		return 0;
	}
        
	fifo_size = fir_driver->SCE.fifo_size - 2;

	/* Transmit current frame */	        
	do {
    /* Avoid failed transmit byte in 16550A, if FIFO get empty and only one 
       byte left to transmit, then that byte may not be transmited until the 
       next frame fifo_size = 2 means at this point that there is only 1 place 
       left in fifo.   We break here, so that there are 2 bytes left in buffer 
       to transmit at next interrupt.
    */
		cnt = driver->tx.count - driver->tx.ptr;
		if ((cnt == 2) && (fifo_size == 2)) {
      DEBUG(0, "irport_write: Breaking...\n");
			break;
		}
		if (cnt <= 0) {

      /* Go in here if we've finished sending the packed. */
      if (count > 0)
      {
			  DEBUG(4, "irport_write: finished with frame\n"); 
			  DEBUG(4, "queue len: %d\n", skb_queue_len( &driver->tx.list));

        /* Unlock transmit buffer */
        if(test_bit(0, &fir_driver->tbusy))
        {
          irda_unlock(&fir_driver->tbusy);
          DEBUG(3, "UNLocking IRPORT\n");
        }
      }

			if ((skb = skb_dequeue(&driver->tx.list))) {
				irport_xmit( driver, skb);
      }

			break;
		}	        

		/* Transmit next byte */
		outb( *driver->tx.head, iobase+UART_TX);
	  //printk("%c", (char) *((char *)driver->tx.head));
		driver->tx.ptr++;
		driver->tx.head++;
		count++;
	} while (--fifo_size > 0);

  if (count > 0)
  {
	  DEBUG(4, "Write finished.  %d bytes written, fifo_size = %d\n", count, fifo_size);
  }

	return count;
}

/*
 * Function irport_xmit (void)
 *
 *    Transmits the current frame until FIFO is full, then
 *    waits until the next transmitt interrupt, and continues until the
 *    frame is transmited.
 */

void irport_xmit(struct driver_cb *driver, struct sk_buff *skb) 
{
  struct fir_driver_cb *fir_driver = (struct fir_driver_cb *)driver->self;
	int iobase = fir_driver->SCE.iobase;
	DEBUG(4, "irport_xmit()\n");

	if ( !driver)
		return;
	        
	if (fir_driver->baud > 115200) {
		DEBUG(4, "irport_xmit: baud=%d, sending to FIR_xmit!\n", fir_driver->baud);
		DEBUG(4, "irport_xmit: skb->len=%ld\n", skb->len);
		memcpy(driver->tx.buff, skb->data, skb->len);
		driver->tx.count = skb->len;
		driver->tx.head = driver->tx.buff;
		driver->tx.ptr = 0;
		FIR_xmit(fir_driver);
	} else {
    /* Check to make sure that there's room to send the frame */
	  if ((inb_p( iobase+UART_LSR) & UART_LSR_THRE) 
        && (driver->tx.count - driver->tx.ptr <= 0) 
        && (!test_bit(0, &fir_driver->sbusy))) {
      driver->tx.count = irport_async_wrap( skb, driver->tx.buff, 
          driver->tx.buffsize); 
      driver->tx.head = driver->tx.buff; 
      driver->tx.ptr = 0;

      DEBUG(3, "Locking IRPORT\n");
      irda_lock( &fir_driver->tbusy);
    } else {
      /* There wasn't room, so queue it for next time */
      DEBUG(0, "Queueing Frame.\n");
      skb_queue_tail(&(driver->tx.list), skb);
      return;
	  }

		if(irport_write(driver) == 0)
    {
      DEBUG(0, "***ERROR*** FIFO is full.  This should never happen. Write failed.");
    }

	}


	IS_SKB( skb, return;);
	FREE_SKB_MAGIC( skb);
	DEV_KFREE_SKB( skb, FREE_WRITE);
}
        
/*
 * Function irport_receive (void)
 *
 *    Reads available bytes from the FIFO, and unwraps into 
 *    frame(s), which are eventually passed up to the upper layer in the
 *    protocol stack 
 *
 *    return: number of bytes received (may not be entire frame)
 *	 	-1 on error
 */
int irport_receive(struct fir_driver_cb *driver) 
{
	__u8 byte = 0x00;
	int iobase;
	int count = 0;

	if ( !driver)
		return -1;

	DEBUG(5, "--> irport_receive\n");
	
	iobase = driver->SCE.iobase;

	/* 
	 *  Receive all characters available in FIFO 
	 */
	do {
		byte = inb_p( iobase+UART_RX);
		irport_async_unwrap(&driver->drv, byte);
		count++;
	} while ( inb_p( iobase+UART_LSR) & UART_LSR_DR);

  DEBUG(4, "irport_receive: got %d bytes\n", count);
 
	return count;
}

/*
 * Function irport_set_media_busy (status)
 *
 *    Called when we have detected that another station is transmitting
 *    in contention mode.
 *
 */
void irport_set_media_busy( struct driver_cb *driver, int status) 
{
	if (driver == NULL)
		return;

	if ( status) {
		driver->media_busy = TRUE;
		irport_start_media_timer(driver);
		DEBUG(4, "Media busy!\n");
	} else {
		driver->media_busy = FALSE;
		del_timer( &driver->media_busy_timer);
	}
}

void irport_start_media_timer( struct driver_cb *driver) 
{
	del_timer( &driver->media_busy_timer);
        
	driver->media_busy_timer.data	  = (unsigned long) driver;
	driver->media_busy_timer.function = &irport_media_busy_expired;
	driver->media_busy_timer.expires  = jiffies + MEDIABUSY_TIMEOUT;
        
	add_timer( &driver->media_busy_timer);
}


/*
 * Function irport_media_busy_expired (data)
 *
 *    
 */
void irport_media_busy_expired(unsigned long data) 
{
	struct driver_cb *driver = (struct driver_cb *) data;

	DEBUG(4, "Mediabusy timer expired!\n");
        
	irport_set_media_busy( driver, FALSE);
}

void irport_start_todo_timer( struct driver_cb *driver, int timeout) 
{
	del_timer( &driver->todo_timer);
        
	driver->todo_timer.data     = (unsigned long) driver;
	driver->todo_timer.function = &irport_todo_expired;
	driver->todo_timer.expires  = jiffies + timeout;
        
	add_timer( &driver->todo_timer);
}

void irport_todo_expired( unsigned long data)
{
  struct driver_cb *driver = ( struct driver_cb *) data;

  DEBUG( 4, __FUNCTION__ "()\n");

  /* Check that we still exist */
  if ( !driver || driver->magic != DRIVER_MAGIC) {
    return;
  }

  ASSERT( driver->change_speed != NULL, return;);

  if ( ((struct fir_driver_cb *)driver->self)->new_speed) {
    driver->change_speed( driver, 
        ((struct fir_driver_cb *)driver->self)->new_speed);
  }
}
int irport_is_media_busy(struct driver_cb *driver) 
{
	return driver->media_busy;
}

int irport_is_tbusy(struct driver_cb *driver) 
{
	return driver->media_busy;
}

/*
 * Function irport_is_receiving (irvdd)
 *
 */
int irport_is_receiving( struct fir_driver_cb *self)
{
	DEBUG(4, "IrPORT: is_receiving()\n");

	if (!self || self->drv.magic != DRIVER_MAGIC) {
		DEBUG(0, "irport_is_receiving: BAD MAGIC!\n");
		return FALSE;
	}

	return 1;
}

/* 
 *   Module stuff 
 */

/*
 * Function cleanup_module (void)
 *
 *    
 *
 */
void cleanup_module(void)
{
	DEBUG(4, "IrPORT: cleanup_module!\n");
	irport_cleanup(irport_drv);
}

/*
 * Function init_module (void)
 *
 *    
 *
 */
int init_module(void)
{
  /* 1 - Initialize the driver */
	if (irport_init() < 0) {
			cleanup_module(); 
		return 1;
	}

	return(0);
}

/*
 * Function irport_bh (driver)
 *
 *    Pass frames up to the higher layer (IRLAP) once they've come in. 
 *
 */

void irport_bh(void *data)
{
  struct driver_cb *driver = (struct driver_cb *) data;
  struct sk_buff *skb = NULL;

  /* Process all received frames */
  while (( skb = skb_dequeue( &driver->rx.list)) != NULL) { 
    DEBUG( 4, "irport_bh: Delivering Frame to irlan_input\n");
    irlap_input(driver->irlap, skb);
  }
}



/*
 * Function irport_async_unwrap (skb)
 *
 *    Parse and de-stuff frame received from the IR-port
 *
 */

void irport_async_unwrap( struct driver_cb *driver, __u8 byte) 
{
        int state = driver->rx.state;
        struct sk_buff *skb;

        DEBUG( 6, "irport_async_unwrap()\n");

        /* State machine for receiving frames */           
        switch( state) {
        case OUTSIDE_FRAME:
          printk("O ");
                if ( byte == BOF) {
                        DEBUG( 4, "BOF found, setting state to BEGIN_FRAME and in_frame to TRUE\n");
                        state = BEGIN_FRAME;
                        driver->rx.in_frame = TRUE;
                } else if ( byte == EOF) {
                        irport_set_media_busy(driver, TRUE);
                }
                break;
        case BEGIN_FRAME:
          printk("B ");
                switch ( byte) {
                case BOF:
                        DEBUG( 4, "XBOF found!\n");
                        /* Continue */
                        break;
                case CE:
                        DEBUG( 4, "CE found!\n");
                        /* Stuffed byte */
                        state = LINK_ESCAPE;
                        break;
                case EOF:
                        /* Abort frame */
                        DEBUG( 0, "Frame abort (1)\n");
                        state = OUTSIDE_FRAME;
                        break;
                default:
                        /* Got first byte of frame */
                        if ( driver->rx.count < driver->tx.buffsize)  {
                                driver->rx.buff[ driver->rx.count++] = byte;
                        
                                driver->rx.fcs = IR_FCS ( INIT_FCS, byte);
                                state = INSIDE_FRAME;
                        } else 
                                printk( "Rx buffer overflow\n");
                        break;
                }
                break;
        case LINK_ESCAPE:
          printk("L ");
                switch ( byte) {
                case BOF:
                        /* New frame? */
                        DEBUG( 4, "New frame?");
                        state = BEGIN_FRAME;
                        driver->rx.count = 0;
                        irport_set_media_busy( driver, TRUE);
                        break;
                case CE:
                        DEBUG( 4, "WARNING: State not defined\n");
                        break;
                case EOF:
                        /* Abort frame */
                        DEBUG( 0, "Abort frame (2)\n");
                        state = OUTSIDE_FRAME;
                        driver->rx.count = 0;
                        break;
                default:
                        /* 
                         *  Stuffed char, complement bit 5 of byte 
                         *  following CE, IrLAP p.114 
                         */
                        byte ^= IR_TRANS;
                        if ( driver->rx.count < driver->rx.buffsize)  {
                                driver->rx.buff[ driver->rx.count++] = byte;
                        
                                driver->rx.fcs = IR_FCS( driver->rx.fcs,
byte);
                                state = INSIDE_FRAME;
                        } else 
                                printk( "Rx buffer overflow\n");
                        break;
                }
                break;
        case INSIDE_FRAME:
          printk("%c ", byte);
                switch ( byte) {
                case BOF:
                        /* New frame? */
                        state = BEGIN_FRAME;
                        driver->rx.count = 0;
                        irport_set_media_busy( driver, TRUE);
                        break;
                case CE:
                        /* Stuffed char */
                        state = LINK_ESCAPE;
                        break;
                case EOF:
                        DEBUG( 3, "EOF found.\n");
                        /* End of frame */
                        state = OUTSIDE_FRAME;
                        driver->rx.in_frame = FALSE;
                        
                        /* 
                         *  Test FCS and deliver frame! 
                         */                     
                        if ( driver->rx.fcs == GOOD_FCS) {
                                skb = dev_alloc_skb(driver->rx.count);
                                if (skb == NULL)  {
                                        printk( KERN_INFO 
                                                "irport_async_unwrap: memory squeeze" 
                                                ", dropping frame.\n");
                                        /* sl->rx_dropped++; */
                                        return;
                                }
                                ALLOC_SKB_MAGIC(skb);
#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)
                                skb->free=1;
#endif
                                ASSERT( driver->rx.count-2 > 0, return;);
                                /*
                                 *  Copy data without CRC
                                 */
                                DEBUG( 4, "Frame len=%d\n",
driver->rx.count-2);
                                skb_put( skb, driver->rx.count-2);
                                memcpy( skb->data, driver->rx.buff, 
                                        driver->rx.count-2);
                                driver->rx.count = 0;
                                skb->mac.raw=skb->data;
                                skb_queue_tail( &(driver->rx.list), skb);

                                /*
                                 *  Now we have something to chew on so inform
                                 *  upper 
                                 *  layers.This will schedule a bottom half
                                 *  routine 
                                 *  which is responsible for delivering all 
                                 *  received frames.
                                 */
                                irport_bh(driver);
                                irport_set_media_busy( driver, FALSE);
                        } else {
                                /* 
                                 *  Wrong CRC, discard frame! 
                                 */
                                DEBUG( 0, "Received frame has wrong CRC\n");
                                irport_set_media_busy( driver, TRUE); 
                        }
                        driver->rx.count = 0;
                        break;
                default:
                        /* Next byte of frame */
                        if ( driver->rx.count < driver->rx.buffsize)  {
                                driver->rx.buff[ driver->rx.count++] = byte;
                                
                                driver->rx.fcs = IR_FCS( driver->rx.fcs,
byte);
                        } else 
                                printk( "Rx buffer overflow\n");
                        break;
                }
                break;
        default:
          printk("W ");
        }
        driver->rx.state = state;
}

/* 
 *   Module stuff 
 */

#define MIN_LENGTH 14

inline static int stuff_byte( __u8 byte, __u8 *buf);

/*
 * Function async_wrap (skb, *tx_buff)
 *
 *    Makes a new buffer with wrapping and stuffing, should check that 
 *    we don't get tx buffer overflow.
 */
int irport_async_wrap( struct sk_buff *skb, __u8 *tx_buff, int buffsize) 
{
        __u8 byte;
        __u16 fcs;
        int i, xbofs;
        int n;

        DEBUG( 6, "async_wrap()\n"); 
        ASSERT( skb != NULL, return 0;);

        /* Initialize variables */
        fcs = INIT_FCS;
        n = 0;

        if ( skb->len > 2048) {
                DEBUG( 0,"async_xmit: Warning size=%d of sk_buff to big!\n", 
                       (int) skb->len);
                IS_SKB( skb, return 0;);
                FREE_SKB_MAGIC( skb);
                DEV_KFREE_SKB( skb, FREE_WRITE);
                return 0;
        }       

        /* Send  XBOF's for required min. turn time */
        xbofs = irport_drv->drv.irlap->xbofs_delay;
        for ( i=0; i<xbofs; i++) {
                tx_buff[n++] = XBOF; /* XBOF */
        }

        /* Send negotiated additional BOFs */
        xbofs = irport_drv->drv.irlap->bofs_count;
        for ( i=0; i<xbofs; i++) {
                tx_buff[n++] = BOF; /* BOF */
        }
        
        /* Start of packet character BOF */
        tx_buff[n++] = BOF;

        /* Insert frame and calc CRC */
        for( i=0; i < skb->len; i++) {
                byte = skb->data[i];
                
                /*
                 *  Check for the possibility of tx buffer overflow. We use
                 *  bufsize-5 since the maximum number of bytes that can be 
                 *  transmitted after this point is 5.
                 */
                if ( n > buffsize-5) {
                        printk( KERN_WARNING 
                                "IrDA Wrapper: TX-buffer overflow!\n");
                        return n;
                }
                n+=stuff_byte( byte, tx_buff+n);
                fcs = IR_FCS( fcs, byte);
        }
        
        /* Insert CRC ( LSB first) */
        fcs = ~fcs;
        n += stuff_byte( (__u8) fcs & 0xff, tx_buff+n);
        n += stuff_byte( (__u8) (fcs >> 8) & 0xff, tx_buff+n);

        tx_buff[n++] = EOF;
        
        DEBUG( 6, "async_wrap() -->\n");
        
        return n;
}
 
/*
 * Function stuff_byte (byte, buf)
 *
 *    Byte stuff one single byte and put the result in buffer pointed to by
 *    buf. The buffer must at all times be able to have two bytes inserted.
 * 
 */
inline static int stuff_byte( __u8 byte, __u8 *buf) 
{
        switch ( byte) {
        case BOF:
        case EOF:
        case CE:
                /* Insert transparently coded */
                buf[0] = CE;                   /* Send link escape */
                buf[1] = byte^IR_TRANS;    /* Complement bit 5 */
                return 2;
                /* break; */
        default:
                 /* Non-special value, no transparency required */
                buf[0] = byte;
                return 1;
                /* break; */
        }
}



