/*********************************************************************
 *                
 * Filename:      irtty.c
 * Version:       0.1
 * Description:   IrDA line discipline implementation
 * Status:        Experimental.
 * Author:        Dag Brattli <dagb@cs.uit.no>
 * Created at:    Tue Dec  9 21:18:38 1997
 * Modified at:   Wed Oct  7 13:11:56 1998
 * Modified by:   Dag Brattli <dagb@cs.uit.no>
 * Sources:       slip.c by Laurence Culhane,   <loz@holmes.demon.co.uk>
 *                          Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
 * 
 *     Copyright (c) 1998 Dag Brattli, All Rights Reserved.
 *      
 *     This program is free software; you can redistribute it and/or 
 *     modify it under the terms of the GNU General Public License as 
 *     published by the Free Software Foundation; either version 2 of 
 *     the License, or (at your option) any later version.
 *  
 *     Neither Dag Brattli nor University of Troms admit liability nor
 *     provide warranty for any of this software. This material is 
 *     provided "AS-IS" and at no charge.
 *     
 ********************************************************************/    

#include <linux/module.h>
#if LINUX_VERSION_CODE > 0x020100 
#include <asm/uaccess.h>
#endif
#include <linux/kernel.h>
#include <linux/tty.h>

#include "irda.h"
#include "ircompat.h"
#include "irmod.h"
#include "irtty.h"
#include "wrapper.h"
#include "irlap.h"
#include "timer.h"
#include "driver.h"

static HASHBIN *irtty = NULL;

static struct tty_ldisc irda_ldisc;

extern void esi_change_speed( struct driver_cb *driver, int baud);

char *driver_name = "irtty";

int irtty_init(void) 
{
	int status;
	
	DEBUG( 4, "--> irtty_init()\n");
	
	irtty = hashbin_new( HB_LOCAL);
	if ( irtty == NULL) {
		printk( KERN_WARNING "IrDA: Can't allocate irtty hashbin!\n");
		return -ENOMEM;
	}

	/* Fill in our line protocol discipline, and register it */
	memset( &irda_ldisc, 0, sizeof( irda_ldisc));

	irda_ldisc.magic = TTY_LDISC_MAGIC;
#if LINUX_VERSION_CODE > LinuxVersionCode(2,1,0)
 	irda_ldisc.name  = "irda";
#endif
	irda_ldisc.flags = 0;
	irda_ldisc.open  = irtty_open;
	irda_ldisc.close = irtty_close;
	irda_ldisc.read  = NULL;
	irda_ldisc.write = NULL;
	irda_ldisc.ioctl = (int (*)(struct tty_struct *, struct file *,
				    unsigned int, unsigned long)) irtty_ioctl;
#if LINUX_VERSION_CODE > LinuxVersionCode(2,1,0)
 	irda_ldisc.poll  = NULL;
#endif
	irda_ldisc.receive_buf  = irtty_receive_buf;
	irda_ldisc.receive_room = irtty_receive_room;
	irda_ldisc.write_wakeup = irtty_write_wakeup;
	
	if (( status = tty_register_ldisc( N_IRDA, &irda_ldisc)) != 0)  {
		printk( KERN_ERR 
			"IrDA: can't register line discipline (err = %d)\n", 
			status);
	}
	
	DEBUG( 4, "irtty_init() -->\n");
	
	return status;
}

/* 
 *  Function irtty_cleanup ( )
 *
 *    Called when the irda module is removed. Here we remove all instances
 *    of the driver, and the master array
 */
void irtty_cleanup(void) 
{
	int ret;
	
	DEBUG( 4, __FUNCTION__ "()\n");
	
	/*
	 *  Unregister tty line-discipline
	 */
	if (( ret = tty_register_ldisc( N_IRDA, NULL))) {
		printk( KERN_ERR 
			"IrTTY: can't unregister line discipline (err = %d)\n",
 			ret);
	}

	/*
	 *  The TTY should care of deallocating the instances by using the
	 *  callback to irtty_close(), therefore we do give any deallocation
	 *  function to hashbin_destroy().
	 */
	hashbin_delete( irtty, NULL);

	DEBUG( 4, "IrTTY: cleanup-->\n");
}

/* 
 *  Function irtty_open(tty)
 *
 *    This function is called by the TTY module when the IrDA line
 *    discipline is called for.  Because we are sure the tty line exists,
 *    we only have to link it to a free IrDA channel.  
 */
int irtty_open( struct tty_struct *tty) 
{
	struct irtty_cb *self;
	
	DEBUG( 4, "--> IrTTY: open()\n");

	ASSERT( tty != NULL, return -EEXIST;);

	/* First make sure we're not already connected. */
	self = (struct irtty_cb *) tty->disc_data;
	if ( self != NULL && self->magic == IRTTY_MAGIC)
		return -EEXIST;
	
	/*
	 *  Allocate new instance of the driver
	 */
	self = kmalloc( sizeof(struct irtty_cb), GFP_KERNEL);
	if ( self == NULL) {
		printk( KERN_ERR "IrDA: Can't allocate memory for "
			"IrDA control block!\n");
		return -ENOMEM;
	}
	memset( self, 0, sizeof(struct irtty_cb));
	
	self->tty = tty;
	tty->disc_data = self;

	sprintf( self->name, "%s%d", tty->driver.name,
		 MINOR(tty->device) - 
		 tty->driver.minor_start +
		 tty->driver.name_base);

	DEBUG( 0, "IrTTY: name=%s\n", self->name);

	/* hashbin_insert( irtty, (QUEUE*) self, 0, self->name); */
	hashbin_insert( irtty, (QUEUE*) self, (int) self, NULL);

	if (tty->driver.flush_buffer) {
		tty->driver.flush_buffer(tty);
	}

	if (tty->ldisc.flush_buffer) {
		tty->ldisc.flush_buffer(tty);
	}
	
	/* Initialize timers */
	init_timer( &self->driver.media_busy_timer);
	init_timer( &self->todo_timer);

	self->magic = IRTTY_MAGIC;

	/*
	 *  Initialize driver
	 */
	self->driver.name = self->name;
	driver_open( &self->driver);
	
	self->driver.driver_flags |= SIR_MODE | IO_PIO;
	self->driver.rx.state = OUTSIDE_FRAME;

	/* 
	 *  Initialize QoS capabilities, we fill in all the stuff that
	 *  we support. Be careful not to place any restrictions on values
	 *  that are not device dependent (such as link disconnect time) so
	 *  this parameter can be set by IrLAP (or the user) instead. DB
	 */
	irda_init_max_qos_capabilies( &self->driver.qos);

	/* The only value we must override it the baudrate */
	self->driver.qos.baud_rate.bits = IR_9600|IR_19200|IR_38400|IR_57600|
		                          IR_115200;
	irda_qos_bits_to_value( &self->driver.qos);

	/* Initialize callbacks */
	self->driver.hard_xmit    = irtty_xmit_buf;
	self->driver.change_speed = irtty_change_speed;
/* 	self->driver.is_receiving = irtty_is_receiving; */
	self->driver.is_tbusy     = irtty_is_tbusy;

	self->driver.self = self;

	driver_set_media_busy( &self->driver, FALSE);

	/* Done.  We have linked the TTY line to a channel. */

	MOD_INC_USE_COUNT;

	return 0;
}

/* 
 *  Function irtty_close ( tty)
 *
 *    Close down a IrDA channel. This means flushing out any pending queues,
 *    and then restoring the TTY line discipline to what it was before it got
 *    hooked to IrDA (which usually is TTY again).  
 */
void irtty_close( struct tty_struct *tty) 
{
	struct irtty_cb *self = (struct irtty_cb *) tty->disc_data;
	
	DEBUG( 4, __FUNCTION__ "()\n");

	/* First make sure we're connected. */
	ASSERT( self != NULL, return;);
	ASSERT( self->magic == IRTTY_MAGIC, return;);

	/* Remove driver */
	driver_close( &self->driver);

	/* Stop timers */
	del_timer( &self->driver.media_busy_timer);
	del_timer( &self->todo_timer);

	/* Stop tty */
	tty->flags &= ~(1 << TTY_DO_WRITE_WAKEUP);
	tty->disc_data = 0;

	self->tty = NULL;
	self->magic = ~IRTTY_MAGIC;
	
	/* hashbin_remove( irtty, 0, self->name); */
	self = hashbin_remove( irtty, (int) self, NULL);
	
	if ( self != NULL)
		kfree( self);

 	MOD_DEC_USE_COUNT;

	DEBUG( 4, "IrTTY: close() -->\n");
}

/* 
 *  Function irtty_change_speed ( self, baud)
 *
 *    Change the speed of the serial port
 *
 */
void irtty_change_speed( struct driver_cb *driver, int baud) 
{
        struct termios old_termios;
	struct irtty_cb *self;
	int cflag;

	DEBUG( 4, __FUNCTION__ "()\n");	

	ASSERT( driver != NULL, return;);
	ASSERT( driver->magic == DRIVER_MAGIC, return;);

	self = (struct irtty_cb *) driver->self;

	ASSERT( self != NULL, return;);
	ASSERT( self->magic == IRTTY_MAGIC, return;);

	DEBUG( 6, "Chars in buffer %d\n", 
	       self->tty->driver.chars_in_buffer( self->tty));

	/*
	 *  If we are transmitting, or if the driver has more bytes to send, 
	 *  then we must wait until the frame is sent before changing the
	 *  speed of the serial port. The test is not quite accurate since the
	 *  fifo may still contain bytes to send. 50 msecs should be enough 
	 *  to make it behave correctly, see calculation below:
	 *  
	 *  23 bytes SNRM * 2 for stuffing + 11 bofs + 2 CRC = 59 bytes
	 *  59 bytes / 9600 bps = 49 ms.
	 *
	 *  Some stations may take 200 ms or more before they can operate on
	 *  a new baudrate. (note 5 IrLAP p. 60)
	 */
	if ( self->tbusy || self->tty->driver.chars_in_buffer( self->tty)) {
		self->new_speed = baud;
		
		irda_start_todo_timer( self, 5); /* 50 msecs */
		return;
	}
	
	DEBUG( 6, "Chars in buffer %d\n", 
	       self->tty->driver.chars_in_buffer( self->tty));

	old_termios = *(self->tty->termios);
	cflag = self->tty->termios->c_cflag;

	cflag &= ~CBAUD;

	DEBUG( 4, "IrTTY: *** Setting speed to %d\n", baud);

	switch( baud) {
	case 1200:
		cflag |= B1200;
		break;
	case 2400:
		cflag |= B2400;
		break;
	case 4800:
		cflag |= B4800;
		break;
	case 19200:
		cflag |= B19200;
		break;
	case 38400:
		cflag |= B38400;
		break;
	case 57600:
		cflag |= B57600;
		break;
	case 115200:
		cflag |= B115200;
		break;
	case 9600:
	default:
		cflag |= B9600;
		break;
	}	

	self->tty->termios->c_cflag = cflag;
	self->tty->driver.set_termios( self->tty, &old_termios);

	self->new_speed = 0;
}


/*
 * Function irtty_ioctl (tty, file, cmd, arg)
 *
 *     The swiss army knife of system calls
 *
 */
int irtty_ioctl( struct tty_struct *tty, void *file, int cmd, void *arg) 
{
	struct irtty_cb *self = (struct irtty_cb *) tty->disc_data;
	int err = 0;
	int size = _IOC_SIZE(cmd);

	DEBUG( 4, __FUNCTION__ "()\n");

	/* First make sure we're connected. */
	if ( !self || self->magic != IRTTY_MAGIC) {
		return -EINVAL;
	}

	if ( _IOC_DIR(cmd) & _IOC_READ)
		err = verify_area( VERIFY_WRITE, (void *) arg, size);
	else if ( _IOC_DIR(cmd) & _IOC_WRITE)
		err = verify_area( VERIFY_READ, (void *) arg, size);
	if ( err)
		return err;
	
	switch(cmd) {
	case TCGETS:
	case TCGETA:
		return n_tty_ioctl( tty, (struct file *) file, cmd, 
				    (unsigned long) arg);
		break;
	case IRTTY_IOCTESI:
		DEBUG( 0, __FUNCTION__ "(): ESI IOCTL!!!!!\n");
		self->driver.change_speed = esi_change_speed;

		/* FIXME: this should be done by the esi implementation */
		self->driver.qos.baud_rate.bits = IR_9600|IR_19200|IR_115200;

		/* FIXME: Use this hack for now! DB */
		ASSERT( self->driver.irlap != NULL, return -1;);
		irlap_apply_default_connection_parameters( self->driver.irlap);
		break;
	default:
		return -ENOIOCTLCMD;
	}
	return 0;
}

/* 
 *  Function irtty_receive_buf( tty, cp, count)
 *
 *    Handle the 'receiver data ready' interrupt.  This function is called
 *    by the 'tty_io' module in the kernel when a block of IrDA data has
 *    been received, which can now be decapsulated and delivered for
 *    further processing 
 */
void irtty_receive_buf( struct tty_struct *tty, const unsigned char *cp, 
			char *fp, int count) 
{
	struct irtty_cb *self = (struct irtty_cb *) tty->disc_data;

	DEBUG( 4, "irda_receive_buf(), %d bytes <%ld>\n", count, jiffies);
	
	ASSERT( self != NULL, return;);
	ASSERT( self->magic == IRTTY_MAGIC, return;);
	
	/* Read the characters out of the buffer */
 	while (count--) {
		/* 
		 *  Characters received with a parity error, etc?
		 */
 		if (fp && *fp++) { 
			DEBUG( 0, "Framing or parity error!\n");
			driver_set_media_busy( &self->driver, TRUE);
			/* sl->rx_errors++; */
 			cp++;
 			continue;
 		}
		/*
		 *  Unwrap and destuff one byte
		 */
		async_unwrap_char( &self->driver, *cp++);
		/* self->rx_over_errors++; */

	}
}

/*
 * Function irtty_xmit_buf (skb)
 *
 *    Transmitt buffer
 *
 */
void irtty_xmit_buf( struct driver_cb *driver, struct sk_buff *skb)
{
	struct irtty_cb *self = (struct irtty_cb *) driver->self;
	int actual = 0;
	int xbofs;

	DEBUG( 4, "irtty_xmit_buf()\n");

	ASSERT( driver != NULL, return;);
	ASSERT( driver->magic == DRIVER_MAGIC, return;);
	ASSERT( driver->irlap != NULL, return;);
	ASSERT( driver->irlap->magic == LAP_MAGIC, return;);
	ASSERT( skb != NULL, return;);

	/* Lock transmit buffer */
	irda_lock( &self->tbusy);

        /*  
	 *  Transfer skb to tx_buff while wrapping, stuffing and making CRC 
	 */
	xbofs = irlap_get_xbofs( driver->irlap);
	DEBUG( 4, "Using %d\n", xbofs);
	ASSERT( xbofs < 164, return;);

        driver->tx.count = async_wrap_skb( skb, driver->tx.buff, 
					   driver->tx.buffsize,
					   xbofs); 

	self->tty->flags |= (1 << TTY_DO_WRITE_WAKEUP);
	
	if ( self->tty->driver.write)
		actual = self->tty->driver.write( self->tty, 0, 
						  driver->tx.buff, 
						  driver->tx.count);

	driver->tx.ptr = actual;
	driver->tx.head = driver->tx.buff + actual;

	DEBUG( 4, "actual=%d\n", actual);

	/* 
	 *  Did we transmit the whole frame? Commented out for now since
	 *  I must check if this optimalization really works. DB.
	 */
#if 0
 	if ((driver->tx.count - driver->tx.ptr) <= 0) {
 		DEBUG( 4, "irtty_xmit_buf: finished with frame!\n");
 		self->tty->flags &= ~(1 << TTY_DO_WRITE_WAKEUP);
 		irda_unlock( &self->tbusy);
 	}
#endif

	IS_SKB( skb, return;);
	FREE_SKB_MAGIC( skb);
	DEV_KFREE_SKB( skb, FREE_WRITE);
}

/*
 * Function irtty_receive_room (tty)
 *
 *    Used by the TTY to find out how much data we can receive at a time
 * 
*/
int irtty_receive_room( struct tty_struct *tty) 
{
	DEBUG( 0, "irda_receive_room()\n");

	return 65536;  /* We can handle an infinite amount of data. :-) */
}

/*
 * Function irtty_write_wakeup (tty)
 *
 *    Called by the driver when there's room for more data.  If we have
 *    more packets to send, we send them here.
 *
 */
void irtty_write_wakeup( struct tty_struct *tty) 
{
	int actual = 0, count;
	struct irtty_cb *self = (struct irtty_cb *) tty->disc_data;
	struct driver_cb *driver;
	struct sk_buff *skb;
	
	DEBUG( 4, "irda_write_wakeup() <%ld>\n", jiffies);
	
	/* 
	 *  First make sure we're connected. 
	 */
	ASSERT( self != NULL, return;);
	ASSERT( self->magic == IRTTY_MAGIC, return;);

	driver = &self->driver;

	/*
	 *  Finished with frame?
	 */
	if ( driver->tx.ptr == driver->tx.count)  {

		/* 
		 *  Now serial buffer is almost free & we can start 
		 *  transmission of another packet 
		 */
		/* self->tx_packets++; */
		DEBUG( 4, "irtty_write_wakeup: finished with frame!\n");

		tty->flags &= ~(1 << TTY_DO_WRITE_WAKEUP);
		irda_unlock( &self->tbusy);

	  	if ((skb = skb_dequeue( &driver->tx.list))) {
			irtty_xmit_buf( driver, skb);
		}
		return;
	}
	/*
	 *  Write data left in transmit buffer
	 */
	count = driver->tx.count - driver->tx.ptr;
	actual = tty->driver.write( tty, 0, driver->tx.head, count);
	driver->tx.ptr += actual;
	driver->tx.head += actual;

	DEBUG( 4, "actual=%d, sent %d\n", actual, count);
}

inline int irtty_is_tbusy( struct driver_cb *driver)
{	
	struct irtty_cb *self = (struct irtty_cb *) driver->self;

	ASSERT( self != NULL, return 0;);
	ASSERT( self->magic == IRTTY_MAGIC, return 0;);

	return self->tbusy;
}

void irda_start_todo_timer( struct irtty_cb *self, int timeout) 
{
	ASSERT( self != NULL, return;);
	ASSERT( self->magic == IRTTY_MAGIC, return;);

	del_timer( &self->todo_timer);
	
	self->todo_timer.data     = (unsigned long) self;
	self->todo_timer.function = &irtty_todo_expired;
	self->todo_timer.expires  = jiffies + timeout;
	
	add_timer( &self->todo_timer);
}

void irtty_todo_expired( unsigned long data)
{
	struct irtty_cb *self = ( struct irtty_cb *) data;

	DEBUG( 4, __FUNCTION__ "()\n");
	
	/* Check that we still exist */
	if ( !self || self->magic != IRTTY_MAGIC) {
		return;
	}

	ASSERT( self->driver.change_speed != NULL, return;);

	if ( self->new_speed) {
		self->driver.change_speed( &self->driver, self->new_speed);
	}
}

/*
 * Function init_module (void)
 *
 *    Initialize IrTTY module
 *
 */
int init_module(void)
{
	irtty_init();
	return(0);
}

/*
 * Function cleanup_module (void)
 *
 *    Cleanup IrTTY module
 *
 */
void cleanup_module(void)
{
	DEBUG( 4, "IrTTY: cleanup_module!\n");
	irtty_cleanup();
}


