/***********************************************************************
 *
 * FIR/MIR driver for the Winbond W83977AF Super I/O chip
 *
 * Author: Paul VanderSpek
 *
 * (c) 1998 Corel Computer Corp.
 *
 */


#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/miscdevice.h>

#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/serial_reg.h> 
#include <linux/ioport.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <linux/malloc.h>

#include "irport.h"
#include "ircompat.h"
#include "irda.h"
#include "w83977af.h"
#include "w83977af_ir.h"
#include "driver.h"

#undef DUMPBANK

/**
	Revision History:

	- PaulV, began translating smc driver to become our w977 driver,
	  things work very differently in w977 land.  joy.

  - Ryan Shillington cleaned up the protocol stack, and fixed up the driver to
    work a little more efficiently.

	Notes:

	- driver->rx.state is no longer used.
	- why are we making any calls at all to irport? 
 */


int FIR_receive_mode(struct fir_driver_cb *driver); 

/*
 * Detect and Setup chip.
 *
 * returns non-negative on success.
 */

int FIR_detect(struct fir_driver_cb *driver) 
{
	int iobase;
	int version;
	
	DEBUG(2, "FIR_detect\n");

	/* enter PnP configuration mode */
	w977_efm_enter();

	w977_select_device(W977_DEVICE_IR);

	/* configure PnP port, IRQ, and DMA channel */
	iobase = driver->SCE.iobase;

	w977_write_reg(0x60, (iobase>>8)&0xFF);
	w977_write_reg(0x61, (iobase)&0xFF);
	w977_write_reg(0x70, 0x06);
	w977_write_reg(0x74, 0x01);
	
	/* set append hardware CRC, enable IR bank selection */	
	w977_write_reg(0xF0, APEDCRC|ENBNKSEL);

	/* activate device */
	w977_write_reg(0x30, 0x01);

	w977_efm_exit();
	
	/* read the Advanced IR ID */
	register_bank(iobase, SET3);
	version =  serial_in(iobase, AUID);
	
	/* should be 0x1? */
	if (0x10 != (version&0xF0))
	{
		DEBUG(0, "Wrong chip version");	
		return -1;
	}
	
	/* switch to advanced mode */
	register_bank(iobase, SET2);
  serial_out(iobase, ADCR1, serial_in(iobase, ADCR1) | ADCR1_ADV_SL);
	
	/* set FIFO size to 32 */
	register_bank(iobase, SET2);
	serial_out(iobase, ADCR2, ADCR2_RXFS32|ADCR2_TXFS32);	
	
	/* set FIFO threshold to TX17, RX16 */
	register_bank(iobase, SET0);	
	serial_out(iobase, FCR, FCR_RXTL10|FCR_TXTL10);
	
	/* ensure that we are in legacy mode */
	register_bank(iobase, SET2);
	serial_out(iobase, ADCR1, serial_in(iobase, ADCR1)&~ADCR1_ADV_SL);

	/* enable UART interrupts */
	register_bank(iobase, SET0);
	serial_out(iobase, HCR, HCR_EN_IRQ);
	
	DEBUG(0,"W83977AF driver loaded. Version: 0x%02x\n", version);
	
	register_bank(iobase, SET0);

	return 0;
}

/*
 * FIR_eom
 *
 */

void FIR_eom(struct fir_driver_cb *driver)
{
  struct driver_cb *drv = &driver->drv;
	struct sk_buff *skb;
	int iobase;

	DEBUG(0, "-->FIR_eom!\n");

	if (driver->direction == RCV) {

		drv->rx.count = drv->rx.buffsize 
		  - get_dma_residue(driver->SCE.dma) - 4;
		
		DEBUG( 2, "received %d bytes\n", drv->rx.count);

		skb = dev_alloc_skb(drv->rx.count);
		if (skb == NULL)  {
			printk( KERN_INFO "async_unwrap: memory squeeze, dropping frame.\n");
			return;
		}

		ALLOC_SKB_MAGIC(skb);
#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)
		skb->free=1;
#endif
		ASSERT( drv->rx.count > 0, return;);
		/*
		 *  Copy data without CRC
		 */
		DEBUG( 4, "Frame len=%d\n", drv->rx.count);
		skb_put( skb, drv->rx.count);
		memcpy( skb->data, drv->rx.buff, drv->rx.count);
		drv->rx.count = 0;
		skb->mac.raw=skb->data;
		skb_queue_tail( &(drv->rx.list), skb);

		/*
		 *  Now we have something to chew on so inform upper 
		 *  layers.This will schedule a bottom half routine 
		 *  which is responsible for delivering all 
		 *  received frames.
		 */
		driver->media_busy = FALSE;
		/** FIXME -- ? irport_set_media_busy(driver, FALSE); */
	}	

	iobase = driver->SCE.iobase;

	/* disable DMA */
	register_bank(iobase, SET0);
	serial_out(iobase, HCR, serial_in(iobase, HCR) & ~HCR_EN_DMA);

	if ((skb = skb_dequeue(&drv->tx.list))) {
	        irport_xmit( drv, skb);
	} else {
		FIR_receive_mode(driver);
	}
}

/* 
 * void setup_dma(driver, buffer, count, mode)
 *
 */

void setup_dma(struct fir_driver_cb *driver, char *buffer, int count, int mode)
{
	unsigned long flags;

	save_flags(flags);
	cli();

	disable_dma(driver->SCE.dma);
	clear_dma_ff(driver->SCE.dma);
	set_dma_mode(driver->SCE.dma, mode);
	set_dma_addr(driver->SCE.dma, virt_to_bus(buffer));
	set_dma_count(driver->SCE.dma, count);
	enable_dma(driver->SCE.dma);

	restore_flags(flags);
}

/*
 * FIR_receive_mode (driver)
 *
 */

int FIR_receive_mode(struct fir_driver_cb *driver) 
{
  struct driver_cb *drv = &driver->drv;
	int iobase = driver->SCE.iobase;

	DEBUG(2, "--> FIR_receive_mode!\n");

	/* disable DMA */
	serial_out(iobase, HCR, serial_in(iobase, HCR)&~HCR_EN_DMA);

	/* clear buffer -- necessary? */
	memset(drv->rx.buff, 0, drv->rx.buffsize);

	setup_dma(driver, drv->rx.buff, drv->rx.buffsize, DMA_MODE_READ);

	driver->media_busy = FALSE;
	driver->direction = RCV;
	drv->rx.head = 0;

	/* choose DMA Rx, DMA Fairness, and Advanced mode */
	register_bank(iobase, SET2);
	serial_out(iobase, ADCR1, (serial_in(iobase, ADCR1)&~ADCR1_D_CHSW)|ADCR1_DMA_F|ADCR1_ADV_SL);
	
	/* set USR status interrupt (includes frame end) */
	register_bank(iobase, SET0);
	serial_out(iobase, ICR, ICR_EHSRI);
	
	/* enable DMA */
	serial_out(iobase, HCR, serial_in(iobase, HCR)|HCR_EN_DMA);

	DEBUG(2, "FIR_receive_mode -->\n");	
	return 0;
}

/*
 * FIR_setspeed(driver, baud)
 *
 */

void FIR_setspeed(struct fir_driver_cb *driver, int baud)
{
	int ir_mode;
	int iobase = driver->SCE.iobase;

	DEBUG(2, "--> FIR_setspeed()\n");

	/* disable interrupts */
	register_bank(iobase, SET0);
	serial_out(iobase, ICR, 0);

	switch (baud) {
	case 576000:
		ir_mode = HCR_MIR_576;
		DEBUG(0, "FIR_setspeed: handling baud of 576000\n");
		break;
	case 1152000:
		ir_mode = HCR_MIR_1152;
		DEBUG(0, "FIR_setspeed: handling baud of 1152000\n");
		break;
	case 4000000:
		ir_mode = HCR_FIR;
		DEBUG(0, "FIR_setspeed: handling baud of 4000000\n");
		break;
	default:
		ir_mode = HCR_FIR;
		DEBUG(0, "FIR_setspeed: unknown baud rate of %d\n", baud);
		break;
	}

	/* ensure advanced mode */
	register_bank(iobase, SET2);
	serial_out(iobase, ADCR1, serial_in(iobase, ADCR1)|ADCR1_ADV_SL);

	/* set appropriate speed mode */
	register_bank(iobase, SET0);
	serial_out(iobase, HCR, ir_mode);

	DEBUG(2, "FIR_setspeed() -->\n");
}

/*
 * FIR_shutdown(driver)
 *
 */

void FIR_shutdown(struct fir_driver_cb *driver)
{
	int iobase;

	DEBUG(2, "FIR_shutdown()\n");

	iobase = driver->SCE.iobase;

	/* turn off advanced mode interrupts if appropriate */
	register_bank(iobase, SET2);

	if (0 != (serial_in(iobase, ADCR1) & ADCR1_ADV_SL)) {
		/* turn off interrupts */
		register_bank(iobase, SET0);
		serial_out(iobase, ICR, 0); 
	}

	/* turn off advanced mode */
	register_bank(iobase, SET2);
	serial_out(iobase, ADCR1, serial_in(iobase, ADCR1)&~ADCR1_ADV_SL);
	
	/** disable UART interrupts */
	serial_out(iobase, HCR, 0);
}

/*
 * Function FIR_xmit (driver)
 *
 * Sets up a DMA transfer to send the current frame.
 * 
 */

int FIR_xmit(struct fir_driver_cb *driver) 
{
  struct driver_cb *drv = &driver->drv;
	int iobase;
	
	DEBUG(2, "FIR_xmit: driver->tx.count=%d, ptr=%d\n", drv->tx.count, drv->tx.ptr);
	
	iobase = driver->SCE.iobase;

	/* disable DMA */
	register_bank(iobase, SET0);
	serial_out(iobase, HCR, serial_in(iobase, HCR)&~HCR_EN_DMA);

	setup_dma(driver, drv->tx.buff, drv->tx.count, DMA_MODE_WRITE);

	driver->media_busy = TRUE;
	driver->direction = XMIT;

	/* choose transmit DMA channel  */ 
	register_bank(iobase, SET2);
	serial_out(iobase, ADCR1, serial_in(iobase, ADCR1)|ADCR1_D_CHSW|ADCR1_DMA_F|ADCR1_ADV_SL);
	
	/* set interrupt on DMA */
	register_bank(iobase, SET0);
	serial_out(iobase, ICR, ICR_EDMAI);
	
	/* enable DMA */
	register_bank(iobase, SET0);
	serial_out(iobase, HCR, serial_in(iobase, HCR)|HCR_EN_DMA);

	DEBUG(2, "FIR_xmit done\n");	
	return 0;
}

/*
 * FIR_interrupt
 *
 */

void FIR_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	int status;
	int iobase;
	struct fir_driver_cb *driver = (struct fir_driver_cb *) dev_id;
	
	DEBUG(2, "FIR_interrupt --> \n");
	
	iobase = driver->SCE.iobase;
	
	register_bank(iobase, SET0);

	status = serial_in(iobase, ISR);

	/* continue processing while DMA or FEND are available */
	while (status & (ISR_DMA_I|ISR_FEND_I))
	{
		/* transmit completed */
		if (status & ISR_DMA_I)
		{
			/* FIX -- check for underrun? */

			struct sk_buff *skb;

			driver->media_busy = FALSE;

			/* dequeue another buffer for xmission */
			if ((skb = skb_dequeue(&driver->drv.tx.list))) {
				irport_xmit(&driver->drv, skb);
			}
		}
		else if (status & ISR_FEND_I)
		{
			/* received a frame */
			FIR_eom(driver);
		}	

		status = serial_in(iobase, ISR);	
	}
	
	DEBUG(2, "FIR_interrupt --> DONE\n");
}

