/*********************************************************************
 *                
 * Filename:      wrapper.c
 * Version:       
 * Description:   IrDA Wrapper layer
 * Status:        Experimental.
 * Author:        Dag Brattli <dagb@cs.uit.no>
 * Created at:    Mon Aug  4 20:40:53 1997
 * Modified at:   Fri Oct 16 14:52:05 1998
 * Modified by:   Dag Brattli <dagb@cs.uit.no>
 * 
 *     Copyright (c) 1998 Dag Brattli <dagb@cs.uit.no>, All Rights Reserved.
 *     
 *     This program is free software; you can redistribute it and/or 
 *     modify it under the terms of the GNU General Public License as 
 *     published by the Free Software Foundation; either version 2 of 
 *     the License, or (at your option) any later version.
 *
 *     Neither Dag Brattli nor University of Troms admit liability nor
 *     provide warranty for any of this software. This material is 
 *     provided "AS-IS" and at no charge.
 *
 ********************************************************************/

#include <linux/skbuff.h>
#include <asm/byteorder.h>

#include "irda.h"
#include "ircompat.h"
#include "wrapper.h"
#include "irtty.h"
#include "crc.h"
#include "irlap.h"
#include "irlap_frame.h"
#include "driver.h"

#define MIN_LENGTH 14

inline static int stuff_byte( __u8 byte, __u8 *buf);

/*
 * Function async_wrap (skb, *tx_buff)
 *
 *    Makes a new buffer with wrapping and stuffing, should check that 
 *    we don't get tx buffer overflow.
 */
int async_wrap_skb( struct sk_buff *skb, __u8 *tx_buff, int buffsize, 
		    int xbofs)
{
	__u8 byte;
	union {
		__u16 value;
		__u8 bytes[2];
	} fcs;

	int i;
	int n;

 	DEBUG( 6, __FUNCTION__ "()\n"); 
	ASSERT( skb != NULL, return 0;);

	/* Initialize variables */
	fcs.value = INIT_FCS;
	n = 0;

	if ( skb->len > 2048) {
		DEBUG( 0,"async_xmit: Warning size=%d of sk_buff to big!\n", 
		       (int) skb->len);

		return 0;
	}
	
	/*
	 *  Send  XBOF's for required min. turn time and for the negotiated
	 *  additional XBOFS
	 */
	for ( i=0; i<xbofs; i++) {
 		tx_buff[n++] = XBOF; 
 	}

	/* Start of packet character BOF */
	tx_buff[n++] = BOF;

	/* Insert frame and calc CRC */
	for( i=0; i < skb->len; i++) {
		byte = skb->data[i];
		
		/*
		 *  Check for the possibility of tx buffer overflow. We use
		 *  bufsize-5 since the maximum number of bytes that can be 
		 *  transmitted after this point is 5.
		 */
		if ( n > buffsize-5) {
			printk( KERN_WARNING 
				"IrDA Wrapper: TX-buffer overflow!\n");
			return n;
		}
		n+=stuff_byte( byte, tx_buff+n);
		fcs.value = IR_FCS( fcs.value, byte);
	}
	
	/* Insert CRC in Little Endian format (LSB first) */
	fcs.value = ~fcs.value;
#ifdef __LITTLE_ENDIAN
	n += stuff_byte( fcs.bytes[0], tx_buff+n);
	n += stuff_byte( fcs.bytes[1], tx_buff+n);
#else ifdef __BIG_ENDIAN
	n += stuff_byte( fcs.bytes[1], tx_buff+n);
	n += stuff_byte( fcs.bytes[0], tx_buff+n);
#endif
	tx_buff[n++] = EOF;
	
	DEBUG( 6, "async_wrap() -->\n");
	
	return n;
}
 
/*
 * Function async_unwrap (skb)
 *
 *    Parse and de-stuff frame received from the IR-port
 *
 */
void async_unwrap_char( struct driver_cb *self, __u8 byte) {
	struct sk_buff *skb;
	
 	DEBUG( 6, "async_unwrap()\n");

	/* State machine for receiving frames */	   
	switch( self->rx.state) {
	case OUTSIDE_FRAME:
		if ( byte == BOF) {
			self->rx.state = BEGIN_FRAME;
			self->rx.in_frame = TRUE;
		} else if ( byte == EOF) {
			driver_set_media_busy( self, TRUE);
		}
		break;
	case BEGIN_FRAME:
		switch ( byte) {
		case BOF:
			/* Continue */
			break;
		case CE:
			/* Stuffed byte */
			self->rx.state = LINK_ESCAPE;
			break;
		case EOF:
			/* Abort frame */
			DEBUG( 0, "Frame abort (1)\n");
			self->rx.state = OUTSIDE_FRAME;
			break;
		default:
			/* Got first byte of frame */
			if ( self->rx.count < self->rx.buffsize)  {
				self->rx.buff[ self->rx.count++] = byte;
			
				self->rx.fcs = IR_FCS ( INIT_FCS, byte);
				self->rx.state = INSIDE_FRAME;
			} else 
				printk( "Rx buffer overflow\n");
			break;
		}
		break;
	case LINK_ESCAPE:
		switch ( byte) {
		case BOF:
			/* New frame? */
			DEBUG( 4, "New frame?\n");
			self->rx.state = BEGIN_FRAME;
			self->rx.count = 0;
			driver_set_media_busy( self, TRUE);
			break;
		case CE:
			DEBUG( 4, "WARNING: State not defined\n");
			break;
		case EOF:
			/* Abort frame */
			DEBUG( 0, "Abort frame (2)\n");
			self->rx.state = OUTSIDE_FRAME;
			self->rx.count = 0;
			break;
		default:
			/* 
			 *  Stuffed char, complement bit 5 of byte 
			 *  following CE, IrLAP p.114 
			 */
			byte ^= IR_TRANS;
			if ( self->rx.count < self->rx.buffsize)  {
				self->rx.buff[ self->rx.count++] = byte;
			
				self->rx.fcs = IR_FCS( self->rx.fcs, byte);
				self->rx.state = INSIDE_FRAME;
			} else 
				printk( "Rx buffer overflow\n");
			break;
		}
		break;
	case INSIDE_FRAME:
		switch ( byte) {
		case BOF:
			/* New frame? */
			DEBUG( 4, "New frame?\n");
			self->rx.state = BEGIN_FRAME;
			self->rx.count = 0;
			driver_set_media_busy( self, TRUE);
			break;
		case CE:
			/* Stuffed char */
			self->rx.state = LINK_ESCAPE;
			break;
		case EOF:
			/* End of frame */
			self->rx.state = OUTSIDE_FRAME;
			self->rx.in_frame = FALSE;
			
			/* 
			 *  Test FCS and deliver frame! 
			 */			
			if ( self->rx.fcs == GOOD_FCS) {
				skb = dev_alloc_skb( self->rx.count);
				if (skb == NULL)  {
					printk( KERN_INFO 
						"async_unwrap: memory squeeze" 
						", dropping frame.\n");
					/* sl->rx_dropped++; */
					return;
				}
				ALLOC_SKB_MAGIC( skb);
#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)
				skb->free=1;
#endif
				ASSERT( self->rx.count-2 > 0, return;);
				/*
				 *  Copy data without CRC
				 */
				DEBUG( 6, "Frame len=%d\n", self->rx.count-2);
				skb_put( skb, self->rx.count-2);
				memcpy( skb->data, self->rx.buff, 
					self->rx.count-2);
				skb->mac.raw=skb->data;

				self->rx.count = 0;
				/* 
				 *  Feed it to IrLAP layer 
				 */
				irlap_input( self->irlap, skb);
			} else {
				/* 
				 *  Wrong CRC, discard frame! 
				 */

				DEBUG( 4, "Received frame has wrong CRC\n");
				driver_set_media_busy( self, TRUE); 
				self->rx.count = 0;
			}			
			break;
		default:
			/* Next byte of frame */
			if ( self->rx.count < self->rx.buffsize)  {
				self->rx.buff[ self->rx.count++] = byte;
				
				self->rx.fcs = IR_FCS( self->rx.fcs, byte);
			} else 
				printk( "Rx buffer overflow\n");
			break;
		}
		break;
	}
}
 
/*
 * Function stuff_byte (byte, buf)
 *
 *    Byte stuff one single byte and put the result in buffer pointed to by
 *    buf. The buffer must at all times be able to have two bytes inserted.
 * 
 */
inline static int stuff_byte( __u8 byte, __u8 *buf) {
	switch ( byte) {
	case BOF:
	case EOF:
	case CE:
		/* Insert transparently coded */
		buf[0] = CE;                   /* Send link escape */
		buf[1] = byte^IR_TRANS;    /* Complement bit 5 */
		return 2;
		/* break; */
	default:
		 /* Non-special value, no transparency required */
		buf[0] = byte;
		return 1;
		/* break; */
	}
}
	 

