/*********************************************************************
 *                
 * Filename:      client_fsm.c
 * Version:       0.1
 * Description:   IrLAN Client FSM (Finite State Machine)
 * Status:        Experimental.
 * Author:        Dag Brattli <dagb@cs.uit.no>
 * Created at:    Sun Aug 31 20:14:37 1997
 * Modified at:   Thu Sep 24 14:40:43 1998
 * Modified by:   Dag Brattli <dagb@cs.uit.no>
 * 
 *     Copyright (c) 1998 Dag Brattli <dagb@cs.uit.no>, All Rights Reserved.
 *     
 *     This program is free software; you can redistribute it and/or 
 *     modify it under the terms of the GNU General Public License as 
 *     published by the Free Software Foundation; either version 2 of 
 *     the License, or (at your option) any later version.
 *
 *     Neither Dag Brattli nor University of Troms admit liability nor
 *     provide warranty for any of this software. This material is 
 *     provided "AS-IS" and at no charge.
 *
 ********************************************************************/

#include <linux/skbuff.h>

#include <irda.h>
#include <ircompat.h>
#include <iriap.h>
#include <irlmp.h>
#include <irttp.h>

#include "irlan_common.h"
#include "client.h"
#include "client_fsm.h"

static int irlan_client_state_idle ( struct irlan_client_cb *self, 
				     IRLAN_CLIENT_EVENT event, 
				     struct sk_buff *skb, 
				     struct irlan_client_info *info);
static int irlan_client_state_query( struct irlan_client_cb *self, 
				     IRLAN_CLIENT_EVENT event, 
				     struct sk_buff *skb, 
				     struct irlan_client_info *info);
static int irlan_client_state_conn ( struct irlan_client_cb *self, 
				     IRLAN_CLIENT_EVENT event, 
				     struct sk_buff *skb, 
				     struct irlan_client_info *info);
static int irlan_client_state_info ( struct irlan_client_cb *self, 
				     IRLAN_CLIENT_EVENT event, 
				     struct sk_buff *skb, 
				     struct irlan_client_info *info);
static int irlan_client_state_media( struct irlan_client_cb *self, 
				     IRLAN_CLIENT_EVENT event, 
				     struct sk_buff *skb, 
				     struct irlan_client_info *info);
static int irlan_client_state_open ( struct irlan_client_cb *self, 
				     IRLAN_CLIENT_EVENT event, 
				     struct sk_buff *skb, 
				     struct irlan_client_info *info);
static int irlan_client_state_wait ( struct irlan_client_cb *self, 
				     IRLAN_CLIENT_EVENT event, 
				     struct sk_buff *skb, 
				     struct irlan_client_info *info);
static int irlan_client_state_arb  ( struct irlan_client_cb *self, 
				     IRLAN_CLIENT_EVENT event, 
				     struct sk_buff *skb, 
				     struct irlan_client_info *info);
static int irlan_client_state_data ( struct irlan_client_cb *self, 
				     IRLAN_CLIENT_EVENT event, 
				     struct sk_buff *skb, 
				     struct irlan_client_info *info);
static int irlan_client_state_close( struct irlan_client_cb *self, 
				     IRLAN_CLIENT_EVENT event, 
				     struct sk_buff *skb, 
				     struct irlan_client_info *info);
static int irlan_client_state_sync ( struct irlan_client_cb *self, 
				     IRLAN_CLIENT_EVENT event, 
				     struct sk_buff *skb, 
				     struct irlan_client_info *info);

char *irlan_client_state[] = {
	"IRLAN_CLIENT_IDLE",
	"IRLAN_CLIENT_QUERY",
	"IRLAN_CLIENT_CONN",
	"IRLAN_CLIENT_INFO",
	"IRLAN_CLIENT_MEDIA",
	"IRLAN_CLIENT_OPEN",
	"IRLAN_CLIENT_WAIT",
	"IRLAN_CLIENT_ARB",
	"IRLAN_CLIENT_DATA",
	"IRLAN_CLIENT_CLOSE",
	"IRLAN_CLIENT_SYNC",
};

static int (*state[])( struct irlan_client_cb *self, IRLAN_CLIENT_EVENT event, 
		       struct sk_buff *skb, struct irlan_client_info *info) = 
{ 
	irlan_client_state_idle,
	irlan_client_state_query,
	irlan_client_state_conn,
	irlan_client_state_info,
	irlan_client_state_media,
	irlan_client_state_open,
	irlan_client_state_wait,
	irlan_client_state_arb,
	irlan_client_state_data,
	irlan_client_state_close,
	irlan_client_state_sync
};

void irlan_client_do_event( struct irlan_client_cb *self, 
			    IRLAN_CLIENT_EVENT event, 
			    struct sk_buff *skb, 
			    struct irlan_client_info *info) 
{
	(*state[ self->state]) ( self, event, skb, info);
}

void irlan_client_next_state( struct irlan_client_cb *self, 
			      IRLAN_CLIENT_STATE state) 
{
	self->state = state;
}

/*
 * Function irlan_client_state_idle (event, skb, info)
 *
 *    IDLE, We are waiting for an indication that there is a provider
 *    available.
 */
static int irlan_client_state_idle( struct irlan_client_cb *self, 
				    IRLAN_CLIENT_EVENT event, 
				    struct sk_buff *skb, 
				    struct irlan_client_info *info) 
{
	DEBUG( 4, "irlan_client_state_idle()\n");

	ASSERT( self != NULL, return -1;);
	
	switch( event) {
	case IRLAN_DISCOVERY_INDICATION:
		/* Get some values from peer IAS */
#if 0
		iriap_getvaluebyclass_request( self->daddr,
					       /*  "PnP", "DeviceID",  */
					       "Device", "DeviceName",
					       irlan_client_get_value_confirm,
					       self);
#else
		iriap_getvaluebyclass_request( self->daddr,
					       "IrLAN", "IrDA:TinyTP:LsapSel",
					       irlan_client_get_value_confirm,
					       self);
#endif
		irlan_client_next_state( self, IRLAN_CLIENT_QUERY);
		break;
	default:
		DEBUG( 4, "irlan_client_state_idle(), Unknown event %d\n", 
		       event);
		break;
	}
	if ( skb) {
		IS_SKB( skb, return -1;);
		FREE_SKB_MAGIC(skb);
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	return 0;
}

/*
 * Function irlan_client_state_query (event, skb, info)
 *
 *    QUERY, We have queryed the remote IAS and is ready to connect
 *    to provider, just waiting for the confirm.
 *
 */
static int irlan_client_state_query( struct irlan_client_cb *self, 
				     IRLAN_CLIENT_EVENT event, 
				     struct sk_buff *skb, 
				     struct irlan_client_info *info) 
{
	DEBUG( 4, "irlan_client_state_query()\n");

	ASSERT( self != NULL, return -1;);
	ASSERT( self->magic == IRLAN_CLIENT_MAGIC, return -1;);
	
	switch( event) {
	case IAS_PROVIDER_AVAIL:
		ASSERT( info != NULL, return -1;);
		ASSERT( self->tsap_control != 0, return -1;);

		/* irlan_client_provider_confirm( self, info->dlsap_sel); */
		self->open_retries = 0;
		
		irttp_connect_request( self->tsap_control, info->dlsap_sel, 
				       self->daddr,
				       NULL);
		irlan_client_next_state( self, IRLAN_CLIENT_CONN);
		break;
	case IAS_PROVIDER_NOT_AVAIL:
		DEBUG( 0, "irlan_state_query: IAS_PROVIDER_NOT_AVAIL\n");
		irlan_client_next_state( self, IRLAN_CLIENT_IDLE);
		break;
	case LMP_DISCONNECT:
	case LAP_DISCONNECT:
		irlan_client_next_state( self, IRLAN_CLIENT_IDLE);
		break;
	default:
		DEBUG( 0, "irlan_client_state_query(), Unknown event %d\n", 
		       event);
		break;
	}
	if ( skb) {
		IS_SKB( skb, return -1;);
		FREE_SKB_MAGIC(skb);
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	
	return 0;
}

/*
 * Function irlan_client_state_conn (event, skb, info)
 *
 *    CONN, We have connected to a provider but has not issued any
 *    commands yet.
 *
 */
static int irlan_client_state_conn( struct irlan_client_cb *self, 
				    IRLAN_CLIENT_EVENT event, 
				    struct sk_buff *skb, 
				    struct irlan_client_info *info) 
{
	DEBUG( 4, "irlan_client_state_conn()\n");
	
	ASSERT( self != NULL, return -1;);
	
	switch( event) {
	case CONNECT_COMPLETE:
		/* Send getinfo cmd */
		irlan_client_get_provider_info( self);
		irlan_client_next_state( self, IRLAN_CLIENT_INFO);
		break;
	case LMP_DISCONNECT:
	case LAP_DISCONNECT:
		irlan_client_next_state( self, IRLAN_CLIENT_IDLE);
		break;
	default:
		DEBUG( 0, "irlan_state_conn(), Unknown event %d\n", event);
		break;
	}
	if ( skb) {
		IS_SKB( skb, return -1;);
		FREE_SKB_MAGIC(skb);
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	
	return 0;
}

/*
 * Function irlan_client_state_info (self, event, skb, info)
 *
 *    INFO, We have issued a GetInfo command and is awaiting a reply.
 */
static int irlan_client_state_info( struct irlan_client_cb *self, 
				    IRLAN_CLIENT_EVENT event, 
				    struct sk_buff *skb, 
				    struct irlan_client_info *info) 
{
	DEBUG( 4, "irlan_client_state_info()\n");

	ASSERT( self != NULL, return -1;);
	
	switch( event) {
	case IRLAN_CLIENT_DATA_INDICATION:
		ASSERT( skb != NULL, return -1;);
	
		irlan_client_extract_params( self, skb);
		
		irlan_client_next_state( self, IRLAN_CLIENT_MEDIA);
		
		irlan_client_get_media_char( self);
		break;
		
	case LMP_DISCONNECT:
	case LAP_DISCONNECT:
		irlan_client_next_state( self, IRLAN_CLIENT_IDLE);
		break;
	default:
		DEBUG( 0, "irlan_state_info, Unknown event %d\n", event);
		break;
	}
	if ( skb) {
		IS_SKB( skb, return -1;);
		FREE_SKB_MAGIC(skb);
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	
	return 0;
}

/*
 * Function irlan_client_state_media (self, event, skb, info)
 *
 *    MEDIA, The irlan_client has issued a GetMedia command and is awaiting a
 *    reply.
 *
 */
static int irlan_client_state_media( struct irlan_client_cb *self, 
				     IRLAN_CLIENT_EVENT event, 
				     struct sk_buff *skb, 
				     struct irlan_client_info *info) 
{
	DEBUG( 4, "irlan_client_state_media()\n");
	
	ASSERT( self != NULL, return -1;);

	switch( event) {
	case IRLAN_CLIENT_DATA_INDICATION:
		irlan_client_extract_params( self, skb);
		irlan_client_open_data_channel( self);
		irlan_client_next_state( self, IRLAN_CLIENT_OPEN);
		break;
	case LMP_DISCONNECT:
	case LAP_DISCONNECT:
		irlan_client_next_state( self, IRLAN_CLIENT_IDLE);
		break;
	default:
		DEBUG( 0, "irlan_client_state_media, Unknown event %d\n", 
		       event);
		break;
	}
	if ( skb) {
		IS_SKB( skb, return -1;);
		FREE_SKB_MAGIC(skb);
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	
	return 0;
}

/*
 * Function irlan_client_state_open (self, event, skb, info)
 *
 *    OPEN, The irlan_client has issued a OpenData command and is awaiting a
 *    reply
 *
 */
static int irlan_client_state_open( struct irlan_client_cb *self, 
				    IRLAN_CLIENT_EVENT event, 
				    struct sk_buff *skb, 
				    struct irlan_client_info *info) 
{
	DEBUG( 4, "irlan_client_state_open()\n");
	
	ASSERT( self != NULL, return -1;);

	switch( event) {
	case IRLAN_CLIENT_DATA_INDICATION:
		irlan_client_extract_params( self, skb);
		
		/*
		 *  Check if we have got the remote TSAP for data 
		 *  communications
		 */
	  	if ( self->data_dtsap != 0)
  			irttp_connect_request( self->tsap_data, 
					       self->data_dtsap, self->daddr,
					       NULL);
		
  		irlan_client_next_state( self, IRLAN_CLIENT_DATA);
		break;

	case LMP_DISCONNECT:
	case LAP_DISCONNECT:
		irlan_client_next_state( self, IRLAN_CLIENT_IDLE);
		break;
	default:
		DEBUG( 0, "irlan_client_state_open, Unknown event %d\n", 
		       event);
		break;
	}
	
	if ( skb) {
		IS_SKB( skb, return -1;);
		FREE_SKB_MAGIC(skb);
		DEV_KFREE_SKB( skb, FREE_READ);
	}

	return 0;
}

/*
 * Function irlan_client_state_wait (self, event, skb, info)
 *
 *    WAIT, The irlan_client is waiting for the local provider to enter the
 *    provider OPEN state.
 *
 */
static int irlan_client_state_wait( struct irlan_client_cb *self, 
				    IRLAN_CLIENT_EVENT event, 
				    struct sk_buff *skb, 
				    struct irlan_client_info *info) 
{
	DEBUG( 4, "irlan_client_state_wait()\n");
	
	ASSERT( self != NULL, return -1;);
	
	switch( event) {
	case LMP_DISCONNECT:
	case LAP_DISCONNECT:
		irlan_client_next_state( self, IRLAN_CLIENT_IDLE);
		break;
	default:
		DEBUG( 0, "irlan_client_state_wait, Unknown event %d\n", 
		       event);
		break;
	}
	if ( skb) {
		IS_SKB( skb, return -1;);
		FREE_SKB_MAGIC(skb);
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	
	return 0;
}

static int irlan_client_state_arb( struct irlan_client_cb *self, 
				   IRLAN_CLIENT_EVENT event, 
				   struct sk_buff *skb, 
				   struct irlan_client_info *info) 
{
	DEBUG( 0, "irlan_client_state_arb(), not implemented!\n");
	
	if ( skb) {
		IS_SKB( skb, return -1;);
		FREE_SKB_MAGIC(skb);
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	return 0;
}

/*
 * Function irlan_client_state_data (self, event, skb, info)
 *
 *    DATA, The data channel is connected, allowing data transfers between
 *    the local and remote machines.
 *
 */
static int irlan_client_state_data( struct irlan_client_cb *self, 
				    IRLAN_CLIENT_EVENT event, 
				    struct sk_buff *skb, 
				    struct irlan_client_info *info) 
{
	DEBUG( 4, "irlan_client_state_data()\n");

	ASSERT( self != NULL, return -1;);
	ASSERT( self->magic == IRLAN_CLIENT_MAGIC, return -1;);

	switch( event) {
	case CONNECT_COMPLETE:
		irlan_client_get_unicast_addr( self);
		irlan_client_open_unicast_addr( self);
		irlan_client_open_broadcast_addr( self); 

		DEBUG( 0, "IrLAN, We are now connected!\n");
		self->connected = TRUE;

		/* irlan_next_state( LAN_DATA); */
		break;
	case IRLAN_CLIENT_DATA_INDICATION:
		irlan_client_extract_params( self, skb);

		/* irlan_client_flow_indication( self->data_tsap, FLOW_START,  */
/* 					self); */
		break;
		
	case LMP_DISCONNECT:
	case LAP_DISCONNECT:
		irlan_client_next_state( self, IRLAN_CLIENT_IDLE);
		break;
	default:
		DEBUG( 0, "irlan_client_state_data, Unknown event %d\n", 
		       event);
		break;
	}
	if ( skb) {
		IS_SKB( skb, return -1;);
		FREE_SKB_MAGIC(skb);
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	
	return 0;
}

/*
 * Function irlan_client_state_close (self, event, skb, info)
 *
 *    
 *
 */
static int irlan_client_state_close( struct irlan_client_cb *self, 
				     IRLAN_CLIENT_EVENT event, 
				     struct sk_buff *skb, 
				     struct irlan_client_info *info) 
{
	DEBUG( 0, "irlan_client_state_close()\n");

	if ( skb) {
		IS_SKB( skb, return -1;);
		FREE_SKB_MAGIC(skb);
		DEV_KFREE_SKB( skb, FREE_READ);
	}

	return 0;
}

/*
 * Function irlan_client_state_sync (self, event, skb, info)
 *
 *    
 *
 */
static int irlan_client_state_sync( struct irlan_client_cb *self, 
				    IRLAN_CLIENT_EVENT event, 
				    struct sk_buff *skb, 
				    struct irlan_client_info *info) 
{
	DEBUG( 0, "irlan_client_state_sync()\n");
	
	if ( skb) {
		IS_SKB( skb, return -1;);
		FREE_SKB_MAGIC(skb);
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	
	return 0;
}

