/*********************************************************************
 *                
 * Filename:      irlan_eth.c
 * Version:       
 * Description:   
 * Status:        Experimental.
 * Author:        Dag Brattli <dagb@cs.uit.no>
 * Created at:    Thu Oct 15 08:37:58 1998
 * Modified at:   Wed Oct 21 01:36:50 1998
 * Modified by:   Dag Brattli <dagb@cs.uit.no>
 * Sources:       skeleton.c by Donald Becker <becker@CESDIS.gsfc.nasa.gov>
 *                slip.c by Laurence Culhane,   <loz@holmes.demon.co.uk>
 *                          Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
 * 
 *     Copyright (c) 1998 Dag Brattli, All Rights Reserved.
 *      
 *     This program is free software; you can redistribute it and/or 
 *     modify it under the terms of the GNU General Public License as 
 *     published by the Free Software Foundation; either version 2 of 
 *     the License, or (at your option) any later version.
 *  
 *     Neither Dag Brattli nor University of Troms admit liability nor
 *     provide warranty for any of this software. This material is 
 *     provided "AS-IS" and at no charge.
 *     
 ********************************************************************/

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/if_arp.h>
#include <net/arp.h>

#include <irda.h>
#include <ircompat.h>
#include "irlan.h"
#include "irlan_eth.h"

/*
 * Function irlan_eth_init (dev)
 *
 *    The network device initialization function. Called only once.
 *
 */
int irlan_eth_init( struct device *dev)
{
	struct irlan_cb *self;

	DEBUG( 4, __FUNCTION__"()\n");

	ASSERT( dev != NULL, return -1;);
       
	self = (struct irlan_cb *) dev->priv;

/*  	dev->open               = irlan_eth_open;  */
/* 	dev->stop	        = irlan_eth_close; */

	dev->hard_start_xmit    = irlan_eth_tx; 
	dev->get_stats	        = irlan_eth_get_stats;
	dev->set_multicast_list = irlan_eth_set_multicast_list;

	/* Try to avoid skb_realloc_headroom() */
	dev->hard_header_len = ETH_HLEN + IRLAN_MAX_HEADER;

	dev->tbusy = 1;
	
	ether_setup( dev);
	
	dev->tx_queue_len = TTP_MAX_QUEUE;

	return 0;
}

/*
 * Function irlan_eth_open (dev)
 *
 *    Start the IrLAN ether network device, this function will be called by
 *    "ifconfig irlan0 up".
 *
 */
int irlan_eth_open( struct device *dev)
{
	/* struct irlan_cb *self = (struct irlan_cb *) dev->priv; */
	
	DEBUG( 4, __FUNCTION__ "()\n");

	ASSERT( dev != NULL, return -1;);

	/* Ready to play! */
	dev->tbusy = 0;
	dev->interrupt = 0;
	dev->start = 1;

	/* MOD_INC_USE_COUNT; */
	
	return 0;
}

/*
 * Function irlan_eth_close (dev)
 *
 *    Stop the Client ether network device, his function will be called by
 *    ifconfig down.
 */
int irlan_eth_close(struct device *dev)
{
	DEBUG( 4, __FUNCTION__ "()\n");
	
	ASSERT( dev != NULL, return -1;);
	
	/* Stop device */
	dev->tbusy = 1;
	dev->start = 0;

	/* MOD_DEC_USE_COUNT; */
	
	return 0;
}

/*
 * Function irlan_eth_tx (skb)
 *
 *    Transmits ethernet frames over IrDA link.
 *
 */
int irlan_eth_tx( struct sk_buff *skb, struct device *dev)
{
	struct irlan_cb *self;
#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)
	struct sk_buff *new_skb;
	int headroom;
#endif
	
	DEBUG( 4, __FUNCTION__ "()\n");
	
	self = (struct irlan_cb *) dev->priv;

	ASSERT( self != NULL, return 0;);
	ASSERT( self->magic == IRLAN_MAGIC, return 0;);

	if ( dev->tbusy) {
		/*
		 * If we get here, some higher level has decided we are broken.
		 * There should really be a "kick me" function call instead.
		 */
		int tickssofar = jiffies - dev->trans_start; 
		DEBUG( 4, __FUNCTION__ "(), tbusy==TRUE\n");
		
		if ( tickssofar < 5) 
 			return -EBUSY;
		
 		dev->tbusy = 0;
 		dev->trans_start = jiffies;
	}
	/*
	 * If some higher layer thinks we've missed an tx-done interrupt
	 * we are passed NULL. Caution: dev_tint() handles the cli()/sti()
	 * itself.
	 */
	if ( skb == NULL) {
		DEBUG( 0, __FUNCTION__ "(), skb==NULL\n");
#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)
		dev_tint(dev); 
#endif
		return 0;
	}
	/*
	 *  Check that we are connected
         */
	ALLOC_SKB_MAGIC( skb);
	if ( !self->connected) {
		DEBUG( 0, __FUNCTION__ "(), Not connected, dropping frame!\n");

                IS_SKB( skb, return 0;);
                FREE_SKB_MAGIC( skb);
		DEV_KFREE_SKB( skb, FREE_WRITE);
                ++self->stats.tx_dropped;
		
                return 0;
	}
	
	/*
	 * Block a timer-based transmit from overlapping. This could better be
	 * done with atomic_swap(1, dev->tbusy), but set_bit() works as well.
	 */
	if ( test_and_set_bit(0, (void*) &dev->tbusy) != 0) {
 		printk( KERN_WARNING "%s: Transmitter access conflict.\n", 
			dev->name);
		return 0;
	}	
	DEBUG( 4, "Room left at head: %d\n", skb_headroom(skb));
	DEBUG( 4, "Room left at tail: %d\n", skb_tailroom(skb));
	DEBUG( 4, "Required room: %d\n", IRLAN_MAX_HEADER);
	
	/* Skb headroom large enough to contain IR-headers? */
#if LINUX_VERSION_CODE > LinuxVersionCode(2,1,0)
	if (( skb_headroom( skb) < IRLAN_MAX_HEADER) || ( skb_shared( skb))) {
		struct sk_buff *new_skb = 
			skb_realloc_headroom(skb, IRLAN_MAX_HEADER);
		ASSERT( new_skb != NULL, return 0;);
		ASSERT( skb_headroom( new_skb) >= IRLAN_MAX_HEADER, return 0;);
		ALLOC_SKB_MAGIC( new_skb);
#else
	headroom = skb_headroom( skb);
	if (  headroom < IRLAN_MAX_HEADER) {
		new_skb = dev_alloc_skb( skb->truesize+IRLAN_MAX_HEADER-
 					 headroom-sizeof(struct sk_buff));
		/* new_skb = dev_alloc_skb( skb->len+IRLAN_MAX_HEADER); */
		if ( new_skb == NULL) {
			DEBUG( 0, __FUNCTION__
			       "(), Could not allocate an skb of length %d\n",
			       2048);
			return 0;
		}
		ALLOC_SKB_MAGIC( new_skb);
		new_skb->free=1;
		
		/* 
		 *  Reserve space for TTP, LMP, and LAP header 
		 */
		ASSERT( skb_headroom( new_skb) >= IRLAN_MAX_HEADER, return 0;);
		skb_reserve( new_skb, IRLAN_MAX_HEADER);
		
		/*  
		 *  Set skb->len and skb->tail, and copy buffer to
		 *  transmitt buffer
		 */
		skb_put( new_skb, skb->len);
		memcpy( new_skb->data, skb->data, skb->len);

		DEBUG( 4, "new_skb->len = %ld\n", new_skb->len);
#endif
		/* 
		 *  Free original skb, and use the new one 
		 */
		FREE_SKB_MAGIC( skb);
		DEV_KFREE_SKB( skb, FREE_WRITE);
		skb = new_skb;
	} 

	dev->trans_start = jiffies;
	self->stats.tx_packets++;
#if LINUX_VERSION_CODE > LinuxVersionCode(2,1,0)
	self->stats.tx_bytes += skb->len; 
#endif
	/*
	 *  Now queue the packet in the transport layer
	 *  FIXME: clean up the code below! DB
	 */
	if ( self->use_udata) {
		irttp_udata_request( self->tsap_data, skb);
		dev->tbusy = 0;
	
		return 0;
	}

	if ( irttp_data_request( self->tsap_data, skb) == -1) {
		/*  
		 *  IrTTPs tx queue is full, so we just have to drop the
		 *  frame! You might think that we should just return -1
		 *  and don't deallocate the frame, but that is dangerous
		 *  since it's possible that we have replaced the original
		 *  skb with a new one with larger headroom, and that would
		 *  really confuse do_dev_queue_xmit() in dev.c! I have
		 *  tried :-) DB
		 */
		DEBUG( 4, __FUNCTION__ "(), Dropping frame\n");
		FREE_SKB_MAGIC( skb);
		DEV_KFREE_SKB( skb, FREE_WRITE);
		++self->stats.tx_dropped;
		
		return 0;
	}
	dev->tbusy = 0;
	
	return 0;
}

/*
 * Function irlan_eth_rx (handle, skb)
 *
 *    This function gets the data that is received on the data channel
 *
 */
void irlan_eth_rx( void *instance, void *sap, struct sk_buff *skb)
{
	struct irlan_cb *self;
	
	DEBUG( 4, __FUNCTION__ "()\n");
	
	self = ( struct irlan_cb *) instance;

	ASSERT( self != NULL, return;);
	ASSERT( self->magic == IRLAN_MAGIC, return;);

	if (skb == NULL) {
		++self->stats.rx_dropped; 
		return;
	}
	IS_SKB( skb, return;);
	ASSERT( skb->len > 1, return;);
		

	DEBUG( 4, "Got some ether data: length=%d\n", (int)skb->len); 
	
	skb->dev = &self->dev;
	skb->protocol=eth_type_trans( skb, skb->dev);
	
	/* 
	 *  We haven't deallocated it, but it is out of our control 
	 *  so we just pretend we have deallocated it :-)
	 */
	FREE_SKB_MAGIC( skb);
	
	netif_rx( skb);   /* Eat it! */
	
	self->stats.rx_packets++;
#if LINUX_VERSION_CODE > LinuxVersionCode(2,1,0)
	self->stats.rx_bytes += skb->len; 
#endif	
}

/*
 * Function irlan_eth_flow (status)
 *
 *    Do flow control between IP/Ethernet and IrLAN/IrTTP. This is done by 
 *    controlling the dev->tbusy variable.
 */
void irlan_eth_flow_indication( void *instance, void *sap, LOCAL_FLOW flow)
{
	struct irlan_cb *self;
	struct device *dev;

	DEBUG( 4, __FUNCTION__ "()\n");

	self = (struct irlan_cb *) instance;

	ASSERT( self != NULL, return;);
	ASSERT( self->magic == IRLAN_MAGIC, return;);
	
	dev = &self->dev;

	ASSERT( dev != NULL, return;);
	
	switch ( flow) {
	case FLOW_STOP:
		DEBUG( 4, "IrLAN, stopping Ethernet layer\n");

		dev->tbusy = 1;
		break;
	case FLOW_START:
		/* 
		 *  Tell upper layers that its time to transmit frames again
		 */
		DEBUG( 4, "IrLAN, starting Ethernet layer\n");

		dev->tbusy = 0;

		/* 
		 *  Ready to receive more frames, so schedule the network
		 *  layer
		 */
		mark_bh( NET_BH);		
		break;
	default:
		DEBUG( 0, __FUNCTION__ "(), Unknown flow command!\n");
	}
}

/*
 * Function irlan_eth_rebuild_header (buff, dev, dest, skb)
 *
 *    If we don't want to use ARP. Currently not used!!
 *
 */
void irlan_eth_rebuild_header( void *buff, struct device *dev, 
			       unsigned long dest, struct sk_buff *skb)
{
	struct ethhdr *eth = ( struct ethhdr *) buff;

	memcpy( eth->h_source, dev->dev_addr, dev->addr_len);
	memcpy( eth->h_dest, dev->dev_addr, dev->addr_len);

	/* return 0; */
}

/*
 * Function set_multicast_list (dev)
 *
 *    Configure the filtering of the device
 *
 */
void irlan_eth_set_multicast_list( struct device *dev) 
{
 	struct irlan_cb *self;

 	self = dev->priv; 

 	ASSERT( self != NULL, return;); 
 	ASSERT( self->magic == IRLAN_MAGIC, return;);

	if (dev->mc_count || (dev->flags & IFF_ALLMULTI)) {
		/* Multicast Mode */
		DEBUG( 0, "Multicast mode\n");
		/* irlan_open_broadcast_addr( self); */
	} else if (dev->flags & IFF_PROMISC) {
		/* Promiscous mode */
		DEBUG( 0, "Promiscous mode\n");
	} else {
		/* Normal mode */
		DEBUG( 0, "Normal mode\n");
		/* irlan_open_unicast_addr( self); */
/* 		irlan_open_broadcast_addr( self); */
	}
}

/*
 * Function irlan_get_stats (dev)
 *
 *    Get the current statistics for this device
 *
 */
struct enet_statistics *irlan_eth_get_stats( struct device *dev) 
{
	struct irlan_cb *self = (struct irlan_cb *) dev->priv;

	ASSERT( self != NULL, return NULL;);
	ASSERT( self->magic == IRLAN_MAGIC, return NULL;);

	return &self->stats;
}
