/*********************************************************************
 *                
 * Filename:      server_fsm.c
 * Version:       0.1
 * Description:   IrLAN Server FSM (Finite State Machine)
 * Status:        Experimental.
 * Author:        Dag Brattli <dagb@cs.uit.no>
 * Created at:    Sun Aug 31 20:14:37 1997
 * Modified at:   Tue Sep  8 09:35:14 1998
 * Modified by:   Dag Brattli <dagb@cs.uit.no>
 * 
 *     Copyright (c) 1998 Dag Brattli <dagb@cs.uit.no>, All Rights Reserved.
 *     
 *     This program is free software; you can redistribute it and/or 
 *     modify it under the terms of the GNU General Public License as 
 *     published by the Free Software Foundation; either version 2 of 
 *     the License, or (at your option) any later version.
 *
 *     Neither Dag Brattli nor University of Troms admit liability nor
 *     provide warranty for any of this software. This material is 
 *     provided "AS-IS" and at no charge.
 *
 ********************************************************************/

#include "irda.h"
#include "ircompat.h"
#include "iriap.h"
#include "irlmp.h"
#include "irttp.h"
#include "server.h"
#include "server_fsm.h"

static int irlan_server_state_idle ( struct irlan_server_cb *self, 
				     IRLAN_SERVER_EVENT event, 
				     struct sk_buff *skb, 
				     struct irlan_server_info *info);
static int irlan_server_state_info ( struct irlan_server_cb *self, 
				     IRLAN_SERVER_EVENT event, 
				     struct sk_buff *skb, 
				     struct irlan_server_info *info);
static int irlan_server_state_open ( struct irlan_server_cb *self, 
				     IRLAN_SERVER_EVENT event, 
				     struct sk_buff *skb, 
				     struct irlan_server_info *info);
static int irlan_server_state_data ( struct irlan_server_cb *self, 
				     IRLAN_SERVER_EVENT event, 
				     struct sk_buff *skb, 
				     struct irlan_server_info *info);

char *irlan_server_state[] = {
	"IRLAN_SERVER_IDLE",
	"IRLAN_SERVER_INFO",
	"IRLAN_SERVER_OPEN",
	"IRLAN_SERVER_DATA",
};

static int (*state[])( struct irlan_server_cb *self, IRLAN_SERVER_EVENT event, 
		       struct sk_buff *skb, struct irlan_server_info *info) = 
{ 
	irlan_server_state_idle,
	irlan_server_state_info,
	irlan_server_state_open,
	irlan_server_state_data,
};

void irlan_server_do_event( struct irlan_server_cb *self, 
			    IRLAN_SERVER_EVENT event, 
			    struct sk_buff *skb, 
			    struct irlan_server_info *info) 
{
	(*state[ self->state]) ( self, event, skb, info);
}

void irlan_server_next_state( struct irlan_server_cb *self, 
			      IRLAN_SERVER_STATE state) 
{
	self->state = state;
}

/*
 * Function irlan_server_state_idle (event, skb, info)
 *
 *    IDLE, We are waiting for an indication that there is a provider
 *    available.
 */
static int irlan_server_state_idle( struct irlan_server_cb *self, 
				    IRLAN_SERVER_EVENT event, 
				    struct sk_buff *skb, 
				    struct irlan_server_info *info) 
{
	DEBUG( 4, "irlan_server_state_idle()\n");
	
	ASSERT( self != NULL, return -1;);
	
	switch( event) {
	case CONNECT_INDICATION:
	     ASSERT( info != NULL, return 0;);
	     irlan_server_connect_response( self, info->tsap);
	     irlan_server_next_state( self, IRLAN_SERVER_INFO);
	     break;
	default:
		DEBUG( 4, "irlan_server_state_idle(), Unknown event %d\n", 
		       event);
		break;
	}
	if ( skb) {
		IS_SKB( skb, return -1;);
		FREE_SKB_MAGIC(skb);
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	return 0;
}

/*
 * Function irlan_server_state_info (self, event, skb, info)
 *
 *    INFO, We have issued a GetInfo command and is awaiting a reply.
 */
static int irlan_server_state_info( struct irlan_server_cb *self, 
				    IRLAN_SERVER_EVENT event, 
				    struct sk_buff *skb, 
				    struct irlan_server_info *info) 
{
	DEBUG( 4, "irlan_server_state_info()\n");
	
	ASSERT( self != NULL, return -1;);

	switch( event) {
	case GET_INFO_CMD: 
		irlan_server_send_reply( self, GET_INFO_CMD);
		/* Keep state */
		break;

	case GET_MEDIA_CMD: 
		irlan_server_send_reply( self, GET_MEDIA_CMD);
		/* Keep state */
		break;
		
	case OPEN_DATA_CMD: 
		irlan_server_send_reply( self, OPEN_DATA_CMD);
		irlan_server_next_state( self, IRLAN_SERVER_OPEN);
		break;
	case LMP_DISCONNECT:
	case LAP_DISCONNECT:
		irlan_server_next_state( self, IRLAN_SERVER_IDLE);
		break;
	default:
		DEBUG( 0, "irlan_server_state_info, Unknown event %d\n", 
		       event);
		break;
	}
	if ( skb) {
		IS_SKB( skb, return -1;);
		FREE_SKB_MAGIC(skb);
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	
	return 0;
}

/*
 * Function irlan_server_state_open (self, event, skb, info)
 *
 *    OPEN, The client has issued a OpenData command and is awaiting a
 *    reply
 *
 */
static int irlan_server_state_open( struct irlan_server_cb *self, 
				    IRLAN_SERVER_EVENT event, 
				    struct sk_buff *skb, 
				    struct irlan_server_info *info) 
{
	DEBUG( 4, "irlan_server_state_open()\n");

	ASSERT( self != NULL, return -1;);

	switch( event) {
	case FILTER_CONFIG_CMD:
		irlan_server_extract_params( self, skb);
		irlan_server_send_reply( self, FILTER_CONFIG_CMD);
		/* Keep state */
		break;
		
	case DATA_CONNECT_INDICATION: 
		DEBUG( 4, "DATA_CONNECT_INDICATION\n");
		irlan_server_next_state( self, IRLAN_SERVER_DATA);
		irlan_server_connect_response( self, info->tsap);
		self->connected = TRUE;
		break;
		
	case LMP_DISCONNECT:
	case LAP_DISCONNECT:
		irlan_server_next_state( self, IRLAN_SERVER_IDLE);
		break;
	default:
		DEBUG( 0, "irlan_server_state_open, Unknown event %d\n", 
		       event);
		break;
	}
	
	if ( skb) {
		IS_SKB( skb, return -1;);
		FREE_SKB_MAGIC(skb);
		DEV_KFREE_SKB( skb, FREE_READ);
	}

	return 0;
}

/*
 * Function irlan_server_state_data (self, event, skb, info)
 *
 *    DATA, The data channel is connected, allowing data transfers between
 *    the local and remote machines.
 *
 */
static int irlan_server_state_data( struct irlan_server_cb *self, 
				    IRLAN_SERVER_EVENT event, 
				    struct sk_buff *skb, 
				    struct irlan_server_info *info) 
{
	DEBUG( 4, "irlan_server_state_data()\n");

	ASSERT( self != NULL, return -1;);
	ASSERT( self->magic == IRLAN_SERVER_MAGIC, return -1;);

	switch( event) {
	case FILTER_CONFIG_CMD:
		irlan_server_extract_params( self, skb);
		irlan_server_send_reply( self, FILTER_CONFIG_CMD);
		self->connected = TRUE;
		break;
		
	case LMP_DISCONNECT:
	case LAP_DISCONNECT:
		irlan_server_next_state( self, IRLAN_SERVER_IDLE);
		break;
	default:
		DEBUG( 0, "irlan_server_state_data, Unknown event %d\n", 
		       event);
		break;
	}
	if ( skb) {
		IS_SKB( skb, return -1;);
		FREE_SKB_MAGIC(skb);
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	
	return 0;
}










