/*********************************************************************
 *                
 * Filename:      client_fsm.c<2>
 * Version:       0.1
 * Sources:       irlpt_event.c
 * 
 *     Copyright (c) 1997, Dag Brattli <dagb@cs.uit.no>, All Rights Reserved.
 *     Copyright (c) 1998, Thomas Davis, <ratbert@radiks.net>, All Rights Reserved.
 *     
 *     This program is free software; you can redistribute it and/or 
 *     modify it under the terms of the GNU General Public License as 
 *     published by the Free Software Foundation; either version 2 of 
 *     the License, or (at your option) any later version.
 *
 *     I, Thomas Davis, provide no warranty for any of this software. This 
 *     material is provided "AS-IS" and at no charge.
 *
 ********************************************************************/

#include "ircompat.h"
#include "iriap.h"
#include "irlmp.h"
#include "irttp.h"
#include "common.h"
#include "client.h"
#include "client_fsm.h"
#include "irda.h"

#if 0
static char *rcsid = "$Id: client_fsm.c,v 1.1.1.1 1998/09/29 02:27:14 root Exp $";
#endif

static int client_state_idle  ( struct irlpt_cb *self, IRLPT_EVENT event,
				struct sk_buff *skb, struct irlpt_info *info);
static int client_state_query ( struct irlpt_cb *self, IRLPT_EVENT event,
				struct sk_buff *skb, struct irlpt_info *info);
static int client_state_ready  ( struct irlpt_cb *self, IRLPT_EVENT event, 
				 struct sk_buff *skb, struct irlpt_info *info);
static int client_state_waiti ( struct irlpt_cb *self, IRLPT_EVENT event,
				struct sk_buff *skb, struct irlpt_info *info);
static int client_state_waitr ( struct irlpt_cb *self, IRLPT_EVENT event,
				struct sk_buff *skb, struct irlpt_info *info);
static int client_state_conn  ( struct irlpt_cb *self, IRLPT_EVENT event, 
				struct sk_buff *skb, struct irlpt_info *info);

static int (*state[])( struct irlpt_cb *self, IRLPT_EVENT event, 
		       struct sk_buff *skb, struct irlpt_info *info) = 
{ 
	client_state_idle,
	client_state_query,
	client_state_ready,
	client_state_waiti,
	client_state_waitr,
	client_state_conn,
};

void client_do_event( struct irlpt_cb *self, IRLPT_EVENT event, 
		      struct sk_buff *skb, struct irlpt_info *info) {
	DEBUG( 4, "client_do_event: STATE = %s, EVENT = %s\n", 
	       fsm_state[self->state], fsm_event[event]);
	(*state[ self->state]) ( self, event, skb, info);
}

void client_next_state( struct irlpt_cb *self, IRLPT_STATE state) {
	DEBUG( 4, "client_next_state: NEXT STATE = %s\n", fsm_state[state]);
	self->state = state;
}

/*
 * Function client_state_idle (event, skb, info)
 *
 *    IDLE, We are waiting for an indication that there is a provider
 *    available.
 */
static int client_state_idle( struct irlpt_cb *self, IRLPT_EVENT event, 
			     struct sk_buff *skb, 
			     struct irlpt_info *info) {

	DEBUG( 4, "client_state_idle()\n");

	switch( event) {
	case IRLPT_DISCOVERY_INDICATION:
		/* Get some values from peer IAS */
		DEBUG( 4, "client_state_idle: IRLPT_DISCOVERY_INDICATION\n");
		DEBUG( 4, "sending getvaluebyclass command..\n");
		iriap_getvaluebyclass_request( info->daddr, 
					"IrLPT", "IrDA:IrLMP:LsapSel",
					client_get_value_confirm,
					self);
		client_next_state( self, LPT_QUERY);
		break;

	default:
		DEBUG( 0, "client_state_idle(), Unknown event %d\n", event);
		break;
	}

	if (skb) {
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	return 0;
}

/*
 * Function client_state_query
 *
 *    QUERY, We have queryed the remote IAS and is ready to connect
 *    to provider, just waiting for the confirm.
 *
 */
static int client_state_query( struct irlpt_cb *self, IRLPT_EVENT event, 
			      struct sk_buff *skb, struct irlpt_info *info) {
	DEBUG( 4, "client_state_query()\n");

	switch( event) {
	case IAS_PROVIDER_AVAIL:
		DEBUG( 4, "client_state_query: IAS_PROVIDER_AVAIL\n");
		self->open_retries = 0;
		client_next_state( self, LPT_READY);
		client_do_event( self, LPT_CONNECT_REQUEST, NULL, NULL);
		break;

	case IAS_PROVIDER_NOT_AVAIL:
		DEBUG( 4, "client_state_query: IAS_PROVIDER_NOT_AVAIL\n");
		client_next_state( self, LPT_IDLE);
		break;

	default:
		DEBUG( 0, "client_state_query:, Unknown event %d\n", event);
		break;
	}
	if (skb) {
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	return 0;
}

/*
 * Function client_state_info 
 *
 *    INFO, We have issued a GetInfo command and is awaiting a reply.
 */
static int client_state_ready( struct irlpt_cb *self, IRLPT_EVENT event,
			     struct sk_buff *skb, struct irlpt_info *info) {
	
	DEBUG( 4, "client_state_ready()\n");

	switch( event) {
	case LPT_CONNECT_REQUEST:
	        DEBUG( 4, "client_state_ready: LPT_CONNECT_REQUEST\n");
		client_connect_request(self);
		client_next_state( self, LPT_WAITI);
		break;
	case LMP_DISCONNECT:
	case LAP_DISCONNECT:
	        DEBUG( 4, "client_state_ready: LMP_DISCONNECT or LAP_DISCONNECT\n");
		client_next_state( self, LPT_IDLE);
		break;
	default:
		DEBUG( 0, "client_state_ready: Unknown event %d\n", event);
		break;
	}
	if ( skb) {
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	return 0;
}


/*
 * Function client_state_waiti 
 *
 *
 */
static int client_state_waiti( struct irlpt_cb *self, IRLPT_EVENT event, 
			      struct sk_buff *skb, struct irlpt_info *info) {
	DEBUG( 4, "client_state_waiti()\n");

	switch( event) {
	case LMP_CONNECT:
	        DEBUG( 4, "client_state_waiti: LMP_CONNECT\n");
	        client_next_state(self, LPT_CONN);
	        break;
	case LAP_DISCONNECT:
	case LMP_DISCONNECT:
	        DEBUG( 4, "client_state_waiti: LMP_DISCONNECT\n");
	        client_next_state( self, LPT_IDLE);
	        break;

	default:
		DEBUG( 4, "client_state_waiti: Unknown event %d\n", event);
		break;
	}

	if ( skb) {
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	return 0;
}

/*
 * Function client_state_waitr 
 *
 *
 */
static int client_state_waitr( struct irlpt_cb *self, IRLPT_EVENT event,
			      struct sk_buff *skb, struct irlpt_info *info) {
	DEBUG( 3, "client_state_waitr()\n");

	switch( event) {
	case LMP_CONNECT:
		DEBUG( 3, "client_state_waitr: LMP_CONNECT\n");
		client_next_state(self, LPT_CONN);
		break;

	case LMP_DISCONNECT:
		DEBUG( 3, "client_state_waitr: LMP_DISCONNECT\n");
		client_next_state( self, LPT_IDLE);
		break;

	case LAP_DISCONNECT:
		DEBUG( 3, "client_state_waitr: LAP_DISCONNECT\n");
		client_next_state( self, LPT_IDLE);
		break;

	default:
		DEBUG( 3, "client_state_waitr: Unknown event %d, (%s)\n", event, fsm_event[event]);
		break;
	}

	if ( skb) {
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	return 0;
}

/*
 * Function client_state_conn (event, skb, info)
 *
 *    CONN, We have connected to a provider but has not issued any
 *    commands yet.
 *
 */
static int client_state_conn( struct irlpt_cb *self, IRLPT_EVENT event, 
			     struct sk_buff *skb, struct irlpt_info *info) 
{
	DEBUG( 4, "client_state_conn()\n");

	switch( event) {
	case CLIENT_DATA_INDICATION:
		DEBUG( 4, "client_state_conn: CLIENT_DATA_INDICATION\n");
		client_next_state( self, LPT_CONN);
		break;

	case LMP_DISCONNECT:
	case LAP_DISCONNECT:
		DEBUG( 4, "client_state_conn: LMP_DISCONNECT/LAP_DISCONNECT\n");
		client_next_state( self, LPT_IDLE);
		break;

	default:
		DEBUG( 4, "client_state_conn: Unknown event %d\n", event);
		break;
	}

	if ( skb) {
		DEV_KFREE_SKB( skb, FREE_READ);
	}

	return 0;
}

void client_print_event( IRLPT_EVENT event) {
	DEBUG( 0, "IRLPT_EVENT = %s\n", fsm_event[event]);
}


