/*********************************************************************
 *                
 * Filename:      irlpt_event.c
 * Version:       0.1
 * Sources:       irlan_event.c
 * 
 *     Copyright (c) 1997, Dag Brattli <dagb@cs.uit.no>, All Rights Reserved.
 *     Copyright (c) 1998, Thomas Davis, <ratbert@radiks.net>, All Rights Reserved.
 *     
 *     This program is free software; you can redistribute it and/or 
 *     modify it under the terms of the GNU General Public License as 
 *     published by the Free Software Foundation; either version 2 of 
 *     the License, or (at your option) any later version.
 *
 *     I, Thomas Davis, provide no warranty for any of this software. This 
 *     material is provided "AS-IS" and at no charge.
 *
 ********************************************************************/

#include "ircompat.h"
#include "iriap.h"
#include "irlmp.h"
#include "irttp.h"
#include "irlpt.h"
#include "irlpt_event.h"
#include "irda.h"

static int irlpt_state_idle  ( struct irlpt_cb *self, IRLPT_EVENT event,
			       struct sk_buff *skb, struct irlpt_frame *frame);
static int irlpt_state_query ( struct irlpt_cb *self, IRLPT_EVENT event,
			       struct sk_buff *skb, struct irlpt_frame *frame);
static int irlpt_state_ready  ( struct irlpt_cb *self, IRLPT_EVENT event, 
			       struct sk_buff *skb, struct irlpt_frame *frame);
static int irlpt_state_waiti ( struct irlpt_cb *self, IRLPT_EVENT event,
			       struct sk_buff *skb, struct irlpt_frame *frame);
static int irlpt_state_waitr ( struct irlpt_cb *self, IRLPT_EVENT event,
			       struct sk_buff *skb, struct irlpt_frame *frame);
static int irlpt_state_conn  ( struct irlpt_cb *self, IRLPT_EVENT event, 
			       struct sk_buff *skb, struct irlpt_frame *frame);

#if 0
static char *rcsid = "$Id: irlpt_event.c,v 1.1.1.1 1998/09/29 02:27:15 root Exp $";
#endif

char *irlptstate[] = {
	"IDLE",
	"QUERY",
	"READY",
	"WAITI",
	"WAITR",
	"CONN"
};

char *irlptevent[] = {
        "IRLPT_DISCOVERY_INDICATION",
	"QUERY_REMOTE_IAS",
	"IAS_PROVIDER_AVAIL",
	"IAS_PROVIDER_NOT_AVAIL",
	"LAP_DISCONNECT",
	"LMP_CONNECT",
	"LMP_DISCONNECT",
	"LMP_CONNECT_INDICATION",
	"LMP_DISCONNECT_INDICATION",
	"LPT_CONNECT_REQUEST",
	"LPT_DISCONNECT_REQUEST",
	"CLIENT_DATA_INDICATION",
};

static int (*state[])( struct irlpt_cb *self, IRLPT_EVENT event, struct sk_buff *skb,
					   struct irlpt_frame *frame) = 
{ 
	irlpt_state_idle,
	irlpt_state_query,
	irlpt_state_ready,
	irlpt_state_waiti,
	irlpt_state_waitr,
	irlpt_state_conn,
};

void irlpt_do_event( struct irlpt_cb *self, IRLPT_EVENT event, struct sk_buff *skb, 
					 struct irlpt_frame *frame) {
	DEBUG( 4, "irlpt_do_event: STATE = %s, EVENT = %s\n", irlptstate[self->state], irlptevent[event]);
	(*state[ self->state]) ( self, event, skb, frame);
}

void irlpt_next_state( struct irlpt_cb *self, IRLPT_STATE state) {
	DEBUG( 4, "irlpt_next_state: NEXT STATE = %s\n", irlptstate[state]);
	self->state = state;
}

/*
 * Function irlpt_state_idle (event, skb, frame)
 *
 *    IDLE, We are waiting for an indication that there is a provider
 *    available.
 */
static int irlpt_state_idle( struct irlpt_cb *self, IRLPT_EVENT event, 
			     struct sk_buff *skb, 
			     struct irlpt_frame *frame) {

	struct sk_buff *r_skb;

	DEBUG( 4, "irlpt_state_idle()\n");

	switch( event) {
	case IRLPT_DISCOVERY_INDICATION:
		/* Get some values from peer IAS */
		DEBUG( 4, "irlpt_state_idle: IRLPT_DISCOVERY_INDICATION\n");
		DEBUG( 4, "sending getvaluebyclass command..\n");
		iriap_getvaluebyclass_request( frame->daddr, 
					"IrLPT", "IrDA:IrLMP:LsapSel",
					irlpt_get_value_confirm,
					self);
		irlpt_next_state( self, LPT_QUERY);
		break;

	case LMP_CONNECT:
		DEBUG( 3, "irlpt_state_idle: LM_CONNECT, remote lsap=%d\n", frame->dlsap);

		self->peersap = frame->dlsap;

		r_skb = dev_alloc_skb(64);
		if (r_skb == NULL) { 
			DEBUG( 0, "couldn't allocate an sk_buff of length 64\n");
			return 0;
		}
		ALLOC_SKB_MAGIC(r_skb);
#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)
		r_skb->free=1;
#endif
		skb_reserve( r_skb, HEADER_SIZE);
		skb->len = 0;
		irlmp_connect_response(self->handle, r_skb);
		irlpt_next_state( self, LPT_CONN);

		break;

	default:
		DEBUG( 0, "irlpt_state_idle(), Unknown event %d\n", event);
		break;
	}

	if (skb) {
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	return 0;
}

/*
 * Function irlpt_state_query
 *
 *    QUERY, We have queryed the remote IAS and is ready to connect
 *    to provider, just waiting for the confirm.
 *
 */
static int irlpt_state_query( struct irlpt_cb *self, IRLPT_EVENT event, 
			      struct sk_buff *skb, struct irlpt_frame *frame) {
	DEBUG( 4, "irlpt_state_query()\n");

	switch( event) {
	case IAS_PROVIDER_AVAIL:
		DEBUG( 4, "irlpt_state_query: IAS_PROVIDER_AVAIL\n");
		self->open_retries = 0;
		irlpt_next_state( self, LPT_READY);
		irlpt_do_event( self, LPT_CONNECT_REQUEST, NULL, NULL);
		break;

	case IAS_PROVIDER_NOT_AVAIL:
		DEBUG( 4, "irlpt_state_query: IAS_PROVIDER_NOT_AVAIL\n");
		irlpt_next_state( self, LPT_IDLE);
		break;

	default:
		DEBUG( 0, "irlpt_state_query:, Unknown event %d\n", event);
		break;
	}
	if (skb) {
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	return 0;
}

/*
 * Function irlpt_state_info 
 *
 *    INFO, We have issued a GetInfo command and is awaiting a reply.
 */
static int irlpt_state_ready( struct irlpt_cb *self, IRLPT_EVENT event,
			     struct sk_buff *skb, struct irlpt_frame *frame) {
	
	DEBUG( 4, "irlpt_state_ready()\n");

	switch( event) {
	case LPT_CONNECT_REQUEST:
	        DEBUG( 4, "irlpt_state_ready: LPT_CONNECT_REQUEST\n");
		irlpt_connect_request(self);
		irlpt_next_state( self, LPT_WAITI);
		break;
	case LMP_DISCONNECT:
	case LAP_DISCONNECT:
	        DEBUG( 4, "irlpt_state_ready: LMP_DISCONNECT or LAP_DISCONNECT\n");
		irlpt_next_state( self, LPT_IDLE);
		break;
	default:
		DEBUG( 0, "irlpt_state_ready: Unknown event %d\n", event);
		break;
	}
	if ( skb) {
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	return 0;
}


/*
 * Function irlpt_state_waiti 
 *
 *
 */
static int irlpt_state_waiti( struct irlpt_cb *self, IRLPT_EVENT event, 
			      struct sk_buff *skb, struct irlpt_frame *frame) {
	DEBUG( 4, "irlpt_state_waiti()\n");

	switch( event) {
	case LMP_CONNECT:
	        DEBUG( 4, "irlpt_state_waiti: LMP_CONNECT\n");
	        irlpt_next_state(self, LPT_CONN);
	        break;
	case LAP_DISCONNECT:
	case LMP_DISCONNECT:
	        DEBUG( 4, "irlpt_state_waiti: LMP_DISCONNECT\n");
	        irlpt_next_state( self, LPT_IDLE);
	        break;

	default:
		DEBUG( 4, "irlpt_state_waiti: Unknown event %d\n", event);
		break;
	}

	if ( skb) {
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	return 0;
}

/*
 * Function irlpt_state_waitr 
 *
 *
 */
static int irlpt_state_waitr( struct irlpt_cb *self, IRLPT_EVENT event,
			      struct sk_buff *skb, struct irlpt_frame *frame) {
	DEBUG( 3, "irlpt_state_waitr()\n");

	switch( event) {
	case LMP_CONNECT:
		DEBUG( 3, "irlpt_state_waitr: LMP_CONNECT\n");
		irlpt_next_state(self, LPT_CONN);
		break;

	case LMP_DISCONNECT:
		DEBUG( 3, "irlpt_state_waitr: LMP_DISCONNECT\n");
		irlpt_next_state( self, LPT_IDLE);
		break;

	case LAP_DISCONNECT:
		DEBUG( 3, "irlpt_state_waitr: LAP_DISCONNECT\n");
		irlpt_next_state( self, LPT_IDLE);
		break;

	default:
		DEBUG( 3, "irlpt_state_waitr: Unknown event %d, (%s)\n", event, irlptevent[event]);
		break;
	}

	if ( skb) {
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	return 0;
}

/*
 * Function irlpt_state_conn (event, skb, frame)
 *
 *    CONN, We have connected to a provider but has not issued any
 *    commands yet.
 *
 */
static int irlpt_state_conn( struct irlpt_cb *self, IRLPT_EVENT event, 
			     struct sk_buff *skb, struct irlpt_frame *frame) {
	DEBUG( 4, "irlpt_state_conn()\n");

	switch( event) {
	case CLIENT_DATA_INDICATION:
		DEBUG( 4, "irlpt_state_conn: CLIENT_DATA_INDICATION\n");
		irlpt_next_state( self, LPT_CONN);
		break;

	case LMP_DISCONNECT:
	case LAP_DISCONNECT:
		DEBUG( 4, "irlpt_state_conn: LMP_DISCONNECT/LAP_DISCONNECT\n");
		irlpt_next_state( self, LPT_IDLE);
		break;

	default:
		DEBUG( 4, "irlpt_state_conn: Unknown event %d\n", event);
		break;
	}
	if ( skb) {
		DEV_KFREE_SKB( skb, FREE_READ);
	}
	return 0;
}

void irlpt_print_event( IRLPT_EVENT event) {
	DEBUG( 0, "IRLPT_EVENT = %s\n", irlptevent[event]);
}


