/*********************************************************************
 *                
 * Filename:      irmanager.c
 * Version:       0.1
 * Description:   IrDA Manager Daemon
 * Status:        Experimental.
 * Author:        Dag Brattli <dagb@cs.uit.no>
 * Created at:    Fri Oct 16 22:23:08 1998
 * Modified at:   Wed Oct 21 14:25:28 1998
 * Modified by:   Dag Brattli <dagb@cs.uit.no>
 * Sources:
 *
 *     Copyright (c) 1998 Dag Brattli, All Rights Reserved.
 *      
 *     This program is free software; you can redistribute it and/or 
 *     modify it under the terms of the GNU General Public License as 
 *     published by the Free Software Foundation; either version 2 of 
 *     the License, or (at your option) any later version.
 *  
 *     Neither Dag Brattli nor University of Troms admit liability nor
 *     provide warranty for any of this software. This material is 
 *     provided "AS-IS" and at no charge.
 *     
 ********************************************************************/

#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ioctl.h>

#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <assert.h>
#include <signal.h>
#include <asm/ioctls.h>
#include <syslog.h>

#include "irmanager.h"

extern int lookup_dev(char *name);
extern int open_dev(dev_t dev, int mode);
extern void beep(unsigned int ms, unsigned int freq);
extern int execute(char *msg, char *cmd);
extern int execute_on_dev(char *action, char *class, char *dev);

static char *configpath = "/etc/irda";

#define BEEP_TIME 150
#define BEEP_OK   1000
#define BEEP_WARN 2000
#define BEEP_ERR  4000

/*
 * Function start_class (class, dev_name)
 *
 *    
 *
 */
void start_class( char *class, char *dev_name)
{
	int ret;

	printf( "Start %s\n", class);

	ret = execute_on_dev( "start", class , dev_name);
	if (ret)
		beep(BEEP_TIME, BEEP_ERR);
	else
		beep(BEEP_TIME, BEEP_OK);
}

/*
 * Function stop_class (class, devname)
 *
 *    
 *
 */
void stop_class( char *class,char *devname)
{
	int ret;

	printf( "Stop %s\n", class);

	ret = execute_on_dev( "stop", class, devname);
	if (ret)
		beep(BEEP_TIME, BEEP_ERR);
	else
		beep(BEEP_TIME, BEEP_OK);
}

/*
 * Function set_sysctl_param (name, value)
 *
 *    Set parameter <name> to <value> in /proc/sys/net/irda
 *
 */
int set_sysctl_param( char *name, char *value)
{
	char msg[128], cmd[512];

	sprintf(msg, "Setting %s to %s", name, value);
	sprintf(cmd, "echo %s > /proc/sys/net/irda/%s", value, name);

	return execute(msg, cmd);
}

/*
 * Function load_module (name)
 *
 *    Tries to load (modprobe) the module with the given name
 *
 */
int load_module( char *name)
{
	char msg[128], cmd[512];
	int ret;
	
	sprintf(msg, "Trying to load module %s", name);
	sprintf(cmd, "modprobe %s", name);

	ret = execute(msg, cmd);
	if (ret)
		beep(BEEP_TIME, BEEP_ERR);
	else
		beep(BEEP_TIME, BEEP_OK);

	return ret;
}

/*
 * Function main (argc, )
 *
 *    Main program :-)
 *
 */
int main (int argc, char *argv[])
{
	int fd;	
	int minor;
	struct irda_event event;
	int ret;
	int c;
	
	while (( c = getopt( argc, argv, "c:d:")) != -1) {
		switch ( c) {
		case 'c': /* Compression */
			if ( strcmp( optarg, "1") == 0) {
				load_module( "irda_deflate");
				
				set_sysctl_param( "compression", "1");
			}
			break;
		case 'd': /* Discovery */
			if ( strcmp( optarg, "1") == 0) {
				/* User wants to start discovery */
				set_sysctl_param( "discovery", "1");
			}
			break;

		default:
			break;
		} 
	}
	
	if ( chdir( configpath) != 0)
		syslog( LOG_INFO, "chdir to %s failed: %m", configpath);

	minor = lookup_dev( "irda");

	printf( "IrDA device: major=10, minor=%d\n", minor);

	fd = open_dev((10 << 8)+minor,S_IREAD|S_IFCHR);
	if ( fd == -1) {
		perror( "Unable to open device");
 		fprintf( stderr, "Are you sure you have the irda module "
 			 "loaded?\n");
 		exit( -1); 
 	} 

	while (1) {
		ret = read( fd, &event, sizeof( event));

		switch ( event.event) {
		case EVENT_DEVICE_DISCOVERED:
			printf( "Device \"%s\" discovered!\n", event.info);
			printf( "daddr = %08x\n", event.daddr);
			switch ( event.service) {
			case S_LAN:
				load_module( "irlan_client");
				break;
			case S_PRINTER:
				load_module( "irlpt_client");
				break;
			case S_COMM:
				load_module( "ircomm_tty");
				break;
			case S_OBEX:
				load_module( "irobex");
				break;
			default:
				printf( "Sorry not implemented yet!\n");
				break;
			}
			break;
		case EVENT_IRLAN_START:
			start_class( "network", event.devname);
			break;
		case EVENT_IRLAN_STOP:
			stop_class( "network", event.devname);
			break;
		case EVENT_IROBEX_START:
			start_class( "obex", event.devname);
			break;
		case EVENT_IROBEX_STOP:
			stop_class( "obex", event.devname);
			break;	
		default:	
			printf( "Unknown event\n");
			break;
		}
	}
	close(fd);

	return 0;
}

