/*********************************************************************
 *                
 * Filename:      irmod.c
 * Version:       0.3
 * Description:   IrDA module code and some other stuff
 * Status:        Experimental.
 * Author:        Dag Brattli <dagb@cs.uit.no>
 * Created at:    Mon Dec 15 13:55:39 1997
 * Modified at:   Tue Oct 20 08:41:11 1998
 * Modified by:   Dag Brattli <dagb@cs.uit.no>
 * 
 *     Copyright (c) 1997 Dag Brattli, All Rights Reserved.
 *      
 *     This program is free software; you can redistribute it and/or 
 *     modify it under the terms of the GNU General Public License as 
 *     published by the Free Software Foundation; either version 2 of 
 *     the License, or (at your option) any later version.
 *  
 *     Neither Dag Brattli nor University of Troms admit liability nor
 *     provide warranty for any of this software. This material is 
 *     provided "AS-IS" and at no charge.
 *     
 ********************************************************************/

#include <linux/config.h>
#include <linux/module.h> 

#include "ircompat.h"
#include "irda.h"
#include "irlmp.h"
#include "iriap.h"
#include "irttp.h"

struct irda irda;

extern void irda_proc_register(void);
extern void irda_proc_unregister(void);
extern int irda_sysctl_register(void);
extern void irda_sysctl_unregister(void);

static int irda_open( struct inode * inode, struct file *file);
static int irda_ioctl( struct inode *inode, struct file *filp, 
			 unsigned int cmd, unsigned long arg);

#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)
static void irda_close( struct inode *inode, struct file *file);
static int irda_read( struct inode * inode, struct file * file, 
			char * buffer, int count);
static int irda_write( struct inode * inode, struct file * file,
			 const char *buffer, int count);
static int irda_select( struct inode *, struct file *, int, select_table *);

#else
static int irda_close( struct inode *inode, struct file *file);
static ssize_t irda_read( struct file *file, char *buffer, size_t count, 
			    loff_t *noidea);
static ssize_t irda_write( struct file *file, const char *buffer,
			     size_t count, loff_t *noidea);
static u_int irda_poll( struct file *file, poll_table *wait);
#endif

static struct file_operations irda_fops = {
	NULL,	       /* seek */
	irda_read,     /* read */
	irda_write,    /* write */
	NULL,	       /* readdir */
#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)
	irda_select,
#else
	irda_poll,    /* poll */
#endif
	irda_ioctl,   /* ioctl */
	NULL,	       /* mmap */
	irda_open,
	irda_close,
	NULL,
	NULL,          /* fasync */
};

/*
 * Function init_module (void)
 *
 *    Initialize the irda module
 *
 */
int init_module(void) 
{
	DEBUG( 4, "--> IrDA: init_module\n");

        printk( KERN_INFO "Linux IrDA v0.4 Mon Sep 7 20:59:17 1998 (Dag Brattli)\n");

	driver_init();
	irlap_init();
 	irlmp_init();
	iriap_init();
 	irttp_init();
	
#ifdef USE_PROC
	irda_proc_register();
#endif
#ifdef USE_SYSCTL
	irda_sysctl_register();
#endif

	irda.dev.minor = MISC_DYNAMIC_MINOR;
	irda.dev.name = "irda";
	irda.dev.fops = &irda_fops;
	
	misc_register( &irda.dev);

	DEBUG( 4, "IrDA: init_module -->\n");

	return 0;
}

/*
 * Function cleanup_module (void)
 *
 *    Cleanup the irda module
 *
 */
void cleanup_module(void) {

        DEBUG( 4, "--> IrDA: cleanup_module\n");

        misc_deregister( &irda.dev);

#ifdef USE_SYSCTL
	irda_sysctl_unregister();
#endif	

#ifdef USE_PROC
	irda_proc_unregister();
#endif

	irttp_cleanup();
	iriap_cleanup();
	irlap_cleanup(); /* Must be done before irlmp_cleanup()! DB */

	irlmp_cleanup();
	driver_cleanup();

	DEBUG( 4, "IrDA: cleanup_module -->\n");
}

/*
 * Function irda_lock (lock)
 *
 *    Lock variable
 *
 */
inline int irda_lock( int *lock) {
	if ( test_and_set_bit( 0, (void *) lock))  {
             printk("Trying to lock, already locked variable!\n");
	     return -1;
        }  
	return 0;
}

/*
 * Function irda_unlock (lock)
 *
 *    Unlock variable
 *
 */
inline int irda_unlock( int *lock) {
	if ( !test_and_clear_bit( 0, (void *) lock))  {
		printk("Trying to unlock already unlocked variable!\n");
		return -1;
        }
	return 0;
}

/*
 * Function irda_notify_init (notify)
 *
 *    Used for initializing the notify structure
 *
 */
void irda_notify_init( struct notify_t *notify)
{

	notify->data_indication = NULL;
	notify->udata_indication = NULL;
	notify->connect_confirm = NULL;
	notify->connect_indication = NULL;
	notify->disconnect_indication = NULL;
	notify->flow_indication = NULL;
	notify->instance = NULL;
	strncpy( notify->name, "Unknown", NOTIFY_MAX_NAME);
}

#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)
/*
 * Function skb_copy (skb, priority)
 *
 *    skb_copy function fro Linux-2.0.x since the one in the kernel is not
 *    visible to us.
 *
 */
struct sk_buff *skb_copy(struct sk_buff *skb, int priority)
{
        struct sk_buff *n;
        unsigned long offset;

        /*
         *      Allocate the copy buffer
         */
         
        /* IS_SKB(skb, return NULL;); */
        
        n=alloc_skb(skb->truesize-sizeof(struct sk_buff),priority);
        if(n==NULL)
                return NULL;

        /*
         *      Shift between the two data areas in bytes
         */
         
        offset=n->head-skb->head;

        /* Set the data pointer */
        skb_reserve(n,skb->data-skb->head);
        /* Set the tail pointer and length */
        skb_put(n,skb->len);
        /* Copy the bytes */
        memcpy(n->head,skb->head,skb->end-skb->head);
        n->link3=NULL;
        n->list=NULL;
        n->sk=NULL;
        n->when=skb->when;
        n->dev=skb->dev;
        n->h.raw=skb->h.raw+offset;
        n->mac.raw=skb->mac.raw+offset;
        n->ip_hdr=(struct iphdr *)(((char *)skb->ip_hdr)+offset);
        n->saddr=skb->saddr;
        n->daddr=skb->daddr;
        n->raddr=skb->raddr;
        n->seq=skb->seq;
        n->end_seq=skb->end_seq;
        n->ack_seq=skb->ack_seq;
        n->acked=skb->acked;
        memcpy(n->proto_priv, skb->proto_priv, sizeof(skb->proto_priv));
        n->used=skb->used;
        n->free=1;
        n->arp=skb->arp;
        n->tries=0;
        n->lock=0;
        n->users=0;
        n->pkt_type=skb->pkt_type;
        n->stamp=skb->stamp;
        
	ALLOC_SKB_MAGIC( n);
        /* IS_SKB(n, return NULL;); */
        return n;
}
#endif /* LINUX_VERSION_CODE */

#ifdef CHECK_SKB
/*
 * Function skb_check (skb, head, line, file)
 *
 *    Routine to check if a sk_buff is OK:
 *
 */
int skb_check(struct sk_buff *skb, int head, int line, char *file)
{
	DEBUG( 4, "skb_check()\n");
	if (skb == NULL) {
		printk("File: %s Line %d, found a null skb\n", file, line);
		return -1;
	}
		
        if (head) {
                if (skb->magic_debug_cookie != SK_HEAD_SKB) {
                        printk("File: %s Line %d, found a bad skb-head\n",
			       file,line);
                        return -1;
                }
                if (!skb->next || !skb->prev) {
                        printk("skb_check: head without next or prev\n");
                        return -1;
                }
                if (skb->next->magic_debug_cookie != SK_HEAD_SKB
		    && skb->next->magic_debug_cookie != SK_GOOD_SKB) {
                        printk("File: %s Line %d, bad next head-skb member\n",
			       file,line);
                        return -1;
                }
                if (skb->prev->magic_debug_cookie != SK_HEAD_SKB
		    && skb->prev->magic_debug_cookie != SK_GOOD_SKB) {
                        printk("File: %s Line %d, bad prev head-skb member\n",
			       file,line);
                        return -1;
                }
                return 0;
        }
        if (skb->next != NULL && skb->next->magic_debug_cookie != SK_HEAD_SKB
	    && skb->next->magic_debug_cookie != SK_GOOD_SKB) {
                printk("File: %s Line %d, bad next skb member\n",
		       file,line);
                return -1;
        }
        if (skb->prev != NULL && skb->prev->magic_debug_cookie != SK_HEAD_SKB
	    && skb->prev->magic_debug_cookie != SK_GOOD_SKB) {
                printk("File: %s Line %d, bad prev skb member\n",
		       file,line);
                return -1;
	}

	
        if(skb->magic_debug_cookie==SK_FREED_SKB)
        {
                printk("File: %s Line %d, found a freed skb lurking in the undergrowth!\n",
		       file,line);
                printk("skb=%p, real size=%d",
                        skb,skb->truesize);
                return -1;
        }
        if(skb->magic_debug_cookie!=SK_GOOD_SKB)
        {
                printk("File: %s Line %d, passed a non skb!\n", file,line);
                printk("skb=%p, real size=%d",
                        skb,skb->truesize);
                return -1;
        }
        if(skb->head>skb->data)
	{
		printk("File: %s Line %d, head > data !\n", file,line);
		printk("skb=%p, head=%p, data=%p\n",
		       skb,skb->head,skb->data);
		return -1;
        }
        if(skb->tail>skb->end)
        {
                printk("File: %s Line %d, tail > end!\n", file,line);
                printk("skb=%p, tail=%p, end=%p\n",
                        skb,skb->tail,skb->end);
                return -1;
        }
        if(skb->data>skb->tail)
        {
                printk("File: %s Line %d, data > tail!\n", file,line);
                printk("skb=%p, data=%p, tail=%p\n",
		       skb,skb->data,skb->tail);
                return -1;
        }
        if(skb->tail-skb->data!=skb->len)
        {
                printk("File: %s Line %d, wrong length\n", file,line);
                printk("skb=%p, data=%p, end=%p len=%d\n",
		       skb,skb->data,skb->end,(int)skb->len);
                return -1;
        }
        if((unsigned long) skb->end > (unsigned long) skb)
        {
		return 0;
                printk("File: %s Line %d, control overrun\n", file,line);
                printk("skb=%p, end=%p\n",
		       skb,skb->end);
                return 0;
        }
	
        /* Guess it might be acceptable then */
        return 0;
}
#endif

/*
 * Function irda_send_event (event)
 *
 *    Send an event to the user space manager
 *
 */
int irda_send_event( struct irda_event *event)
{
	unsigned long flags;

	DEBUG( 4, __FUNCTION__ "()\n");

	/* Synchronize */
	save_flags( flags);
	cli();
	if ( irda.ebusy) {
		DEBUG( 0, __FUNCTION__ "(), Event busy!\n");
		restore_flags( flags);
		return -1;
	}
	irda.ebusy = TRUE;
	restore_flags( flags);

	/* Copy event */
	memcpy( &irda.event, event, sizeof( struct irda_event));

	/* Wake up IrDA Manager */
	wake_up_interruptible( &irda.wait_queue);

	return 0;
}

static int irda_open( struct inode * inode, struct file *file)
{
	DEBUG( 0, __FUNCTION__ "()\n");

	MOD_INC_USE_COUNT;

	return 0;
}

static int irda_ioctl( struct inode *inode, struct file *filp, 
		       unsigned int cmd, unsigned long arg)
{
	DEBUG( 0, __FUNCTION__ "()\n");

	return 0;
}

#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)
static void irda_close( struct inode *inode, struct file *file)
#else
static int irda_close( struct inode *inode, struct file *file)
#endif
{
	DEBUG( 0, __FUNCTION__ "()\n");

	MOD_DEC_USE_COUNT;

#if LINUX_VERSION_CODE > LinuxVersionCode(2,1,0)
	return 0;
#endif
}

#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)
static int irda_read( struct inode * inode, struct file * file, 
		      char * buffer, int count)
#else
static ssize_t irda_read( struct file *file, char *buffer, size_t count, 
			  loff_t *noidea)
#endif
{
	unsigned long flags;
	int len;

	DEBUG( 4, __FUNCTION__ "()\n");

	/* * Go to sleep and wait for event if there is no event to be read! */
	save_flags( flags);
	cli();
	if ( !irda.ebusy)
		interruptible_sleep_on( &irda.wait_queue);
	restore_flags(flags);
	/*
	 *  Ensure proper reaction to signals, and screen out 
	 *  blocked signals (page 112. linux device drivers)
	 */
#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)
	if ( current->signal & ~current->blocked)
#else
	if ( signal_pending( current))
#endif
		return -ERESTARTSYS;

	len = sizeof(struct irda_event);
	copy_to_user( buffer, &irda.event, len);

	irda.ebusy = 0;
	/* irda_unlock( &irda.ebusy); */

	return len;
}


#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)
static int irda_write( struct inode * inode, struct file * file,
		       const char *buffer, int count)
#else
static ssize_t irda_write( struct file *file, const char *buffer,
			   size_t count, loff_t *noidea)
#endif
{
	DEBUG( 0, __FUNCTION__ "()\n");

	return 0;
}

#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)
static int irda_select( struct inode *inode, struct file *filp, int mode, 
			select_table *table)
{
	DEBUG( 0, __FUNCTION__ "()\n");
	
	if ( mode == SEL_IN) {
		select_wait( &irda.wait_queue, table);
		return 0;
	}
	if ( mode == SEL_OUT) {
		select_wait( &irda.wait_queue, table);
		return 0;
	}
	return 0;
}
#else
static u_int irda_poll( struct file *file, poll_table *wait)
{
	DEBUG( 0, __FUNCTION__ "(), Sorry not implemented yet!\n");

	return 0;
}
#endif
