/*********************************************************************
 *                
 * Filename:      irobex_lib.c
 * Version:       0.2
 * Description:   User space application for IrOBEX
 * Status:        Experimental.
 * Author:        Dag Brattli <dagb@cs.uit.no>
 * Created at:    Fri Jul 17 23:02:02 1998
 * Modified at:   Wed Oct 21 14:41:25 1998
 * Modified by:   Dag Brattli <dagb@cs.uit.no>
 * 
 *     Copyright (c) 1998 Dag Brattli, All Rights Reserved.
 *      
 *     This program is free software; you can redistribute it and/or 
 *     modify it under the terms of the GNU General Public License as 
 *     published by the Free Software Foundation; either version 2 of 
 *     the License, or (at your option) any later version.
 *  
 *     Neither Dag Brattli nor University of Troms admit liability nor
 *     provide warranty for any of this software. This material is 
 *     provided "AS-IS" and at no charge.
 *     
 *
 * Revision History:
 *
 *   16-Sep-98 09:39:47   Dag Brattli <dagb@cs.uit.no>
 *     FIXME: fix some major memory leaks with the IrOBEX object
 *
 ********************************************************************/

#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <netinet/in.h>

#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <assert.h>
#include <signal.h>
#include <asm/ioctls.h>

#include "irobex_lib.h"

IROBEX_HANDLE *async_self = NULL; /* Which instance wants SIGIO */

int irobex_need_more_input( IROBEX_HANDLE *self, __u8 *buf, int max_len);
int irobex_input( IROBEX_HANDLE *self);

static int irobex_state_disconnected( IROBEX_HANDLE *self, IROBEX_EVENT event,
				      __u8 *buf, int len, int final);
static int irobex_state_connecting  ( IROBEX_HANDLE *self, IROBEX_EVENT event,
				      __u8 *buf, int len, int final);
static int irobex_state_ready(        IROBEX_HANDLE *self, IROBEX_EVENT event,
				      __u8 *buf, int len, int final);
static int irobex_state_put         ( IROBEX_HANDLE *self, IROBEX_EVENT event,
				      __u8 *buf, int len, int final);
static int irobex_state_get         ( IROBEX_HANDLE *self, IROBEX_EVENT event,
				      __u8 *buf, int len, int final);

static char *irobex_state[] = {
	"STATE_DISCONNECTED",
	"STATE_CONNECTING",
	"STATE_READY",
	"STATE_PUT",
	"STATE_GET",
	"*** ERROR ***", /* Should never be used! */
};

static char *irobex_event[] = {
	"CONNECT_INDICATION",
	"CONNECT_REQUEST",
	"DISCONNECT_INDICATION",
	"DISCONNECT_REQUEST",
	"GET_REQUEST",
	"GET_INDICATION",
	"PUT_REQUEST",
	"PUT_INDICATION",
	"RESPONSE_INDICATION",
	"*** ERROR ***", /* Should never be used! */
};

static int (*state[])( IROBEX_HANDLE *self, IROBEX_EVENT event,
		       __u8 *buf, int len, int final) = 
{ 
	irobex_state_disconnected,
	irobex_state_connecting,
	irobex_state_ready,
	irobex_state_put,
	irobex_state_get,
};

void irobex_next_state( IROBEX_HANDLE *self, IROBEX_STATE state)
{
	assert( self != NULL);

	DEBUG( 0, "irobex_next_state: %s\n", irobex_state[ state]);
	
	self->state = state;
}

/*
 * Function irobex_do_event (self, event, buf, len, final)
 *
 *    Handle IrOBEX event
 *
 */
void irobex_do_event( IROBEX_HANDLE *self, IROBEX_EVENT event,
		      __u8 *buf, int len, int final)
{
	int ret;

	assert( self != NULL);
	
  	DEBUG( 0, "irobex_do_event: event = %s, state = %s\n", 
	       irobex_event[event], irobex_state[ self->state]); 
	
	ret = (*state[ self->state]) ( self, event, buf, len, final);
}

void print_buf( __u8 *buf, int len)
{
	int i;

	for ( i=0; i<len; i++)
		DEBUG( 0, "%02x", buf[i]);
}

/*
 * Function input_handler (signal)
 *
 *    Called by SIGIO
 *
 */
void input_handler( int signal)
{
	DEBUG( 4, "Got some input!\n");

	irobex_input( async_self);
}

/*
 * Function irobex_init ()
 *
 *    Initialize IrOBEX
 *
 */
IROBEX_HANDLE *irobex_init( OBJECT_INDICATION_CALLBACK callback)
{
	IROBEX_HANDLE *self;
	int fd;
	int oflags;

	self = ( IROBEX_HANDLE*) malloc( sizeof(IROBEX_HANDLE));
	if ( !self) {
		perror( "irobex_init()");
		exit( -1);
	}
	memset( self, 0, sizeof( IROBEX_HANDLE));

	fd = open( "/dev/irobex", O_RDWR, 0);
	if ( fd == -1) {
		perror( "Unable to open device");
		fprintf( stderr, "Are you sure you have the irobex module "
			 "loaded?\n");
		exit( -1);
	}

	/* Register for asynchronous notification */
	signal( SIGIO, &input_handler);
	fcntl( fd, F_SETOWN, getpid());
	oflags = fcntl( 0, F_GETFL);
	fcntl( fd, F_SETFL, oflags | FASYNC);

	async_self = self;

	self->fd_dev = fd;
	self->callback = callback;
	self->state = STATE_DISCONNECTED;
	self->put_state = STATE_PUT_IDLE;

	return self;
}

/*
 * Function unicode_to_char (unicode)
 *
 *    
 *
 */
/* char *unicode_to_char( __u8 *unicode) */
/* { */
	/* for ( u=1,a=0;u<length-3;u+=2,a++) { */

/* 		name[a] = buf[n+u]; */
/* 	} */
/* 	n+=length-3; */
/* } */

/*
 * Function insert_uint_header (buf, identifier, value)
 *
 *    Insert a four byte unsigned integer in network byte order
 *    identifier if the header ID which is supposed to consist of
 *    both the header enconding and type
 */
int insert_uint_header( __u8 *buf, __u8 identifier, __u32 value)
{
	int n = 0;

	DEBUG( 4, "insert_uint_header()\n");

	buf[n++] = identifier;
	*( __u32 *)( buf+n) = htonl( value);
	n += 4;
	
	return n;
}

/*
 * Function extract_uint_header (buf, identifier, value)
 *
 *    TODO: make similar functions for the other coding formats
 *
 */
int extract_uint_header( __u8 *buf, __u8 *identifier, __u32 *value)
{
	int n = 0;

	*identifier = buf[n]; n++;
	*value = ntohs(*(__u32 *)(buf+n)); n+=4;

	return 0;
}

/*
 * Function insert_unicode_header (buf, opcode, text)
 *
 *    Insert a char string into the buffer after converting it to a 
 *    unicode string
 */
int insert_unicode_header( __u8 *buf, __u8 opcode, char *text)
{
	int n = 0;
	int i;
	int len;

	DEBUG( 4, "insert_unicode_header()\n");

	assert( text != NULL);
	len = strlen(text) + 1; /* We must remember to copy the \0 */

	buf[n++] = opcode;

	*( __u16 *)( buf+n) = htons( len*2 + 3);
	n += 2;

	for (i=0; i<len; i++) {
		buf[n++] = 0x00;
		buf[n++] = text[i];
	}
	return n;
}

int extract_unicode_header( __u8 *buf, __u8 *identifier, char *text)
{
	int n = 0;
	char name[64];
	int u,a;
	__u16 length;

	*identifier = buf[n]; n++;

	/* Length of name (Unicode is 2 bytes pr. char */
	length = ntohs(*(__u16 *)(buf+n)); n+=2;
			
	DEBUG( 4, "name-len=%d\n", length);
	
	assert( length < 64*2);
	
	/*
	 *  Translate from Unicode to ASCII
	 */
	for ( u=1,a=0;u<length-3;u+=2,a++) {
		name[a] = buf[n+u];
	}
	n+=length-3;
	DEBUG( 4, "Name=%s\n",name);	

	return 0;
}


/*
 * Function insert_byte_stream_header (buf, opcode, stream, size)
 *
 *    Insert a byte stream into the buffer
 *
 */
int insert_byte_stream_header( __u8 *buf, __u8 opcode, __u8 *stream, 
			       int size)
{
	int n = 0;

	DEBUG( 4, "insert_byte_stream_header()\n");
	
	buf[n++] = opcode;

	*( __u16 *)( buf+n) = htons( size+3);
	n += 2;

	memcpy( buf+n, stream, size);
	n += size;
	
	return n;
}

/*
 * Function irobex_connect_request (self)
 *
 *    Try to connect (using IrOBEX) to remote device
 *
 */
int irobex_connect_request( IROBEX_HANDLE *self)
{
	char buf[7];
	int size;
	int actual;
	int ret;

	DEBUG( 4, "irobex_connect_request()\n");

	assert( self != NULL);

	/*
	 *  Create packet
	 */
	buf[0] = FINAL | CMD_CONNECT;
	buf[1] = 0x00; /* length MSB */
	buf[2] = 0x07; /* length LSB */
	buf[3] = OBEX_VERSION;
	buf[4] = 0x00; /* flags */
	buf[5] = 0x04; /* max packet size (MSB) */
	buf[6] = 0x05; /* max packet size (LSB) */

	size = 7;

	/*
	 *  Open TTP connection to remote device!
	 */
	ret = ioctl( self->fd_dev, IROBEX_IOCSCONNECT, &self->daddr); 
	if ( ret == -1) { 
		perror( "Unable to make TTP connection to remote device!"); 
		return -1;
	}

	/*
	 *  TTP connection open, so now its time to send and IrOBEX connect
	 *  request!
	 */
	/* print_buf( buf, size); */
	actual = write( self->fd_dev, buf, size);
	if ( actual == -1) {
		perror( "IrOBEX, connect_request: Unable to write");
		exit( -1);
	}
	return 0;
}

/*
 * Function irobex_connect_response (self)
 *
 *    Accept connection and give a response to connecting device
 * 
 */
void irobex_connect_response( IROBEX_HANDLE *self)
{
	__u8 buf[7];
	int actual;

	DEBUG( 4, "irobex_connect_response()\n");

	assert( self != NULL);

	/*
	 *  Create packet
	 */
	buf[0] = FINAL| RSP_SUCCESS; /* success */
	buf[1] = 0x00; /* length MSB */
	buf[2] = 0x07; /* length LSB */
	buf[3] = OBEX_VERSION; 
	buf[4] = 0x00; /* flags */
	buf[5] = 0x04; /* max packet size (MSB) */
	buf[6] = 0x05; /* max packet size (LSB) */

	/*
	 * Send it!
	 */
	/* print_buf( buf, 7); */
	actual = write( self->fd_dev, buf, 7);
	assert( actual == 7);
}

/*
 * Function irobex_connect_indication (self, buf, actual)
 *
 *    Received a connect request from a remote device
 *
 */
void irobex_connect_indication( IROBEX_HANDLE *self, __u8 *buf, int actual)
{
	__u16 length;
	__u8 version;
	int flags;
	__u16 max_packet_length;
	int n = 0;

	DEBUG( 4, "irobex_connect_indication()\n");

	DEBUG( 4, __FUNCTION__ "(), length=%d\n", actual);
	
	assert( self != NULL);
	assert( buf[n] == 0x00); n++;
	
	length = ntohs(*(__u16 *)(buf+n)); n+=2;
	DEBUG( 4, "Packet length=%d\n", length);

	assert( length <= actual);

	version = buf[n++];
	DEBUG( 0, "version=%d.%d\n", version >> 4, version & 0x0f);

	flags = buf[n++];
	
	max_packet_length = ntohs(*(__u16 *)(buf+n)); n+=2;
	DEBUG( 0, "max obex packet length=%d\n", max_packet_length);
	self->mtu = max_packet_length;
}

/*
 * Function irobex_disconnect_request (self)
 *
 *    Take down connection with peer device
 *
 */
void irobex_disconnect_request( IROBEX_HANDLE *self)
{
	__u8 buf[3];
	int actual;

	DEBUG( 4, "irobex_disconnect_request()\n");
	
	assert( self != NULL);

	buf[0] = FINAL | CMD_DISCONNECT;
	buf[1] = 0x00; /* length MSB */
	buf[2] = 0x03; /* length LSB */

	/*
	 * Send it!
	 */
	/* print_buf( buf, 3); */
	actual = write( self->fd_dev, buf, 3);
	/* assert( actual == 3); */
	
	/* 
	 *  Do not close the TTP connection until we get 
	 *  the response 
	 */
}

/*
 * Function irobex_disconnect_indication (self, buf, actual, final)
 *
 *    Remote peer device has closed connection
 *
 */
void irobex_disconnect_indication( IROBEX_HANDLE *self, __u8 *buf, int actual,
				   int final)
{
	int n=0;
	int length;

	DEBUG( 4, __FUNCTION__ "()\n");

	assert( self != NULL);
	assert( buf[n] == CMD_DISCONNECT); n++;

	length = ntohs(*(__u16 *)(buf+n)); n+=2;

	DEBUG( 4, "length=%d\n", length);

/* 	if ( self->callback) */
/* 		self->callback( &self->object); */
}

/*
 * Function irobex_put_response (self, final)
 *
 *    Make some progress
 *
 */
void irobex_put_response( IROBEX_HANDLE *self, int response, int final)
{
	__u8 buf[3];
	int actual;

	DEBUG( 4, "irobex_put_response()\n");
	
	assert( self != NULL);

	/*
	 *  Create packet
	 */
	if ( final)
		buf[0] = FINAL | response; /* response and final*/
	else
		buf[0] = response;
	buf[1] = 0x00; /* length MSB */
	buf[2] = 0x03; /* length LSB */

	/*
	 * Send it!
	 */
	/* print_buf( buf, 3); */
	actual = write( self->fd_dev, buf, 3);
	if ( actual != 3) {
		DEBUG( 0, __FUNCTION__ 
		       "(), Unable to send response! Disconnecte?\n");
	}
}

/*
 * Function irobex_response_indication (self, buf, actual)
 *
 *    Check response code
 *
 */
int irobex_response_indication( IROBEX_HANDLE *self, __u8 *buf, int actual)
{
	int response;

	assert( self != NULL);

	DEBUG( 4, "irobex_get_response()\n");

	response = buf[0];
	
	switch( response) {
	case RSP_CONTINUE:
		printf( "Continue\n");
		break;
	case RSP_SWITCH_PRO:
		printf( "Switching protocols\n");
		break;
	case RSP_SUCCESS:
		printf( "OK, success\n");
		break;
	case RSP_CREATED:
		printf( "Created\n");
		break;
	case RSP_ACCEPTED:
		printf( "Accepted\n");
		break;	
	case RSP_CONFLICT:
		printf( "Conflict\n");
		break;
	default:
		printf( "Hmm, unknown response 0x%02x\n", buf[0]);
	}

	return response;
}

/*
 * Function irobex_put_request (self)
 *
 *    Implementation of the PUT command. Sends an object to the remote device
 *
 */
int irobex_put_request( IROBEX_HANDLE *self)
{
	IROBEX_OBJECT *object;
	int n=0;
	int len=0;
	int size_index;
	int finished = 0;
	int actual = 0;
	__u8 buf[4096];

	object = self->object;

	/*
	 *  Connect before starting the PUT command
	 */
	if ( self->state == STATE_DISCONNECTED) {
		irobex_do_event( self, CONNECT_REQUEST, NULL, 0, 0);
		
		self->put_state = STATE_PUT_HEADERS;
		return finished;
	}

	switch ( self->put_state) {
	case STATE_PUT_HEADERS:
		DEBUG( 4, "STATE_PUT_HEADERS\n");
		/*
		 *  Put command
		 */
		buf[n++] = CMD_PUT;
		
		/* 
		 *  Size of this command, since we don't know that yet, we
		 *  just leave the space open and fill it in at the end
		 */
		size_index = n; n+=2;
		
		/*
		 *  Insert some headers
		 */
		n += insert_unicode_header( buf+n, HEADER_NAME, 
					    object->name);
		n += insert_uint_header(    buf+n, HEADER_LENGTH,
					    object->length);
		n += insert_unicode_header( buf+n, HEADER_DESCRIPTION, 
					    object->description);

#ifdef PALM_III
		n += insert_uint_header(    buf+n, HEADER_CREATOR_ID,
					    object->creator_id);
#endif
		/* 
		 *  Finished with this frame, so it's time to fill in the
		 *  size 
		 */
		*(__u16 *)(buf+size_index) = htons( n);

		/*
		 *  Write the first frame
		 */
		DEBUG( 4, "Write headers %d\n", n);
		/* print_buf( buf, n); */
		actual = write( self->fd_dev, buf, n);
		
		self->put_state = STATE_PUT_BODY;
		break;
	case STATE_PUT_BODY:
		DEBUG( 4, "STATE_PUT_BODY\n");


		if ( self->mtu > 4096)
			self->mtu = 4096;
		/*
		 *  Check if this is the final "normal" body frame
		 */
		if (( object->length - object->index) > self->mtu-3) {
			printf( "This is a normal BODY frame!\n");
			self->put_state = STATE_PUT_BODY;
			len = self->mtu-3;
		} else {
			printf( "This is the final BODY frame!\n");
			self->put_state = STATE_PUT_BODY_END;
			len = object->length - object->index;
		}

		buf[n++] = CMD_PUT;
			
		/* Size of this command */
		size_index = n;  n+=2;

		assert( self->mtu > len);
		n += insert_byte_stream_header( buf+n, HEADER_BODY, 
						object->body+
						object->index, 
						len);
		object->index += len;

		/* Insert the size now that we know it :-) */
		*(__u16 *)(buf+size_index) = htons( n);

		DEBUG( 0, "Write body %d bytes\n", n);
		/* print_buf( buf, n); */
		actual = write( self->fd_dev, buf, n);	

		break;	
	case STATE_PUT_BODY_END:
		DEBUG( 4, "STATE_PUT_BODY_END\n");
		/*
		 *  Send a final empty BODY_END frame
		 */
		buf[n++] = FINAL | CMD_PUT;

		/* Size of this command */
		size_index = n;  n+=2;
		
		n += insert_byte_stream_header( buf+n, HEADER_BODY_END, 
						NULL, 
						0);

		/* Insert the size now that we know it :-) */
		*(__u16 *)(buf+1) = htons( n);

		DEBUG( 0, "Write body end!\n");
		/* print_buf( buf, n); */
		actual = write( self->fd_dev, buf, n);
		
		/* Finished, so deallocate object */
		free( object->body);
		free( self->object);
		self->object = NULL;

		self->put_state = STATE_PUT_IDLE;
		finished = 1;
		break;
	default:
		DEBUG( 0, "irobex_put_request() Unknown PUT state!\n");
		break;
	}
	return finished;
}

/*
 * Function irobex_put_indication (self, buf, actual, final)
 *
 *    Received some object headers from the remote device
 *
 */
void irobex_put_indication( IROBEX_HANDLE *self, __u8 *buf, int len, int final)
{
	IROBEX_OBJECT *object;
	char name[64];
	char description[64];
	__u16 size;
	__u16 length;
	__u32 h_length;
	int n = 0;
	int encoding;
	int count;
	int identifier;
	int u,a;
	int actual = 0;

	DEBUG( 4, "irobex_put_indication(), final=%d\n", final);
	
	assert( self != NULL);

	/* Allocate object to be received*/
	if ( self->object == NULL) {
		DEBUG( 0, __FUNCTION__ "(), Allocating new object!\n");
		self->object = malloc( sizeof(IROBEX_OBJECT));
		memset( self->object, 0, sizeof( IROBEX_OBJECT));
	}

	object = self->object;

	/* assert that this is a put command */
	assert( buf[n] == 0x02); n++;

	/* 
	 *  Get the size for this put command 
	 */
	DEBUG( 0, __FUNCTION__ "Len1 of received data %d\n", len);

	size = ntohs(*(__u16 *)(buf+n)); n+=2;
	DEBUG( 0, "size=%d\n", size);

	while ( size > len) {
		DEBUG( 0, __FUNCTION__ 
		       "(), Need more data, size=%d, len=%d!\n", size, len);

		/*
		 *  Looks like the frame have been fragmented, so try to 
		 *  read the rest of the frame before proceeding
		 */
		actual = irobex_need_more_input( self, buf+size, size-len);

		DEBUG( 0, __FUNCTION__ "(), more=%d\n", actual);

    if(actual < 0)
    {
      fprintf(stderr, "The other party is not co-operating. \n");
      _exit(1);
    }
		len += actual;
    
	}

	DEBUG( 0, __FUNCTION__ "Len2 of received data %d\n", len);

	/*
	 *   Get the OBEX headers
	 */
	while( n < size) {
		encoding = buf[n] & 0xc0; /* mask bit 7 & 8 */
		switch ( encoding) {
		case 0x00:
			DEBUG( 1, "Null terminated Unicode text\n");
			break;
		case 0x40:
			DEBUG( 1, "Byte stream\n");
			break;
		case 0x80:
			DEBUG( 1, "1 byte quantity\n");
			break;
		case 0xc0:
			DEBUG( 1, "4 byte quantity\n");
			break;
		default:
			printf( "Unknown encoding!\n");
			break;
		}
		identifier = buf[n]; n++;

		/*
		 *  Find out which header have we received?
		 */
		switch ( identifier) {
		case HEADER_COUNT: 
			/* 
			 *  Count is the number of objects in this session
			 *  FIXME: only used in connect frames, so it should
			 *  not stay here anymore
			 */
			DEBUG( 4, "Count Header\n");

			count = ntohs(*(__u32 *)(buf+n)); n+=4;
			printf( "Count=%d\n", count);
			break;
		case HEADER_NAME: 
			/* 
			 *  Name header, is the name of the object, usually 
			 *  a file name
			 */
			DEBUG( 4, "Name header\n");

			/* Length of name (Unicode is 2 bytes pr. char */
			length = ntohs(*(__u16 *)(buf+n)); n+=2;
			
			DEBUG( 4, "name-len=%d\n", length);

			assert( length < 64*2);

			/*
			 *  Translate from Unicode to ASCII
			 */
			for ( u=1,a=0;u<length-3;u+=2,a++) {
				name[a] = buf[n+u];
			}
			n+=length-3;
			DEBUG( 4, "Name=%s\n",name);

			strncpy( object->name, name, 64);
			break;
		case HEADER_TYPE: 
			DEBUG( 4, "Type header!\n");
			length = ntohs(*(__u16 *)(buf+n)); n+=2;

			strncpy( object->type, buf+n, length-3);
			
			n+=length-3;

			break;
		case HEADER_TIME:
			DEBUG( 4, "Time header!\n");
			length = ntohs(*(__u16 *)(buf+n)); n+=2;
			
			/* for ( a=0;a<length-3;a++) { */
/* 				DEBUG( 0, "%02x", buf[n+a]); */
/* 			} */
			DEBUG( 4, "\n");
			n+=length-3;
			break;
		case HEADER_LENGTH:
			/*
			 *  Length, is the length of the object in bytes 
			 */
			DEBUG( 4, "Length Header\n");

			h_length = ntohl(*(__u32 *)(buf+n)); n+=4;
			DEBUG( 0, "Length=%d\n", h_length);

			object->length = h_length;
			break;
		case HEADER_DESCRIPTION:
			DEBUG( 4, "Description Header\n");
			/* Length of description */
			length = ntohs(*(__u16 *)(buf+n)); n+=2;
			assert( length < 64*2);
			
			/*
			 *  Translate from Unicode to ASCII
			 */
			for ( u=1,a=0;u<length-3;u+=2,a++) {
				description[a] = buf[n+u];
			}
			n+=length-3;
			DEBUG( 4, "Description=%s\n", description);
			strncpy( object->description, description, 64);
			break;
		case HEADER_BODY:
		case HEADER_BODY_END:
			/*  
			 *  Body, is the binary data of the object
			 *  transfered 
			 */
			DEBUG( 4, "Body Header\n");
			length = ntohs(*(__u16 *)(buf+n)); n+=2;
			
      DEBUG(3, "Length = %d\n", length);
			/* Allocated memory for body yet? */
			if ( object->body == NULL) {
				assert( object->length > 0);

				DEBUG( 0, __FUNCTION__ 
				       "(), allocating %d bytes for object body\n", object->length);
				object->body = malloc( object->length);
        memset(object->body, 0, object->length);
				object->index = 0;
			}
			
			/* Copy body data to object */
			assert( object->body != NULL);
			memcpy( object->body + object->index, buf+n, length-3);

			object->index += length-3;
			DEBUG( 0, __FUNCTION__ "(), index=%d\n", 
			       object->index);
      DEBUG( 4, "Data = %s", buf+n);

			n+=length-3;
			
			break;
		case HEADER_CREATOR_ID:
			object->creator_id = *(__u32 *)(buf+n);
			DEBUG( 4, "Creator ID=0x%08x\n", 
			       object->creator_id);

			n+=4;
			break;
		default:
			printf( "Header %02x Not implemented!\n", identifier);
			return;
			break;
		}
	}
}

/*
 * Function irobex_need_more_input (self, buf, max_len)
 *
 *    Used by irobex_put_indication() to get the rest of a frame which has 
 *    been fragmented by the peer IrOBEX layer. When this happens the frame
 *    may be sent as multiple IrTTP frames with no message boundaries intact
 *
 */
int irobex_need_more_input( IROBEX_HANDLE *self, __u8 *buf, int len)
{
	int actual;

	DEBUG( 0, __FUNCTION__ "()\n");

	actual = read( self->fd_dev, buf, len);

	return actual;
}

/*
 * Function irobex_input (self)
 *
 *    Read some input from device and find out which packet it is
 *
 */
int irobex_input( IROBEX_HANDLE *self)
{
	__u8 buf[4096];
	int final;
	int actual;

	DEBUG( 4, "irobex_input()\n");
	
	assert( self != NULL);

	actual = read( self->fd_dev, buf, 2048);

	DEBUG( 4, "irobex_input: got %d bytes\n", actual);

	if ( actual <= 0)
		return actual;

	final = buf[0] & FINAL;   /* Extract final bit */
	buf[0] = buf[0] & ~FINAL; /* Remove final bit */

	switch( buf[0]) {
	case CMD_CONNECT:
		DEBUG( 4, "Got connect command\n");
		irobex_do_event( self, CONNECT_INDICATION, buf, actual, final);
		break;
	case CMD_DISCONNECT:
		DEBUG( 4, "Got disconnect command\n");
		irobex_do_event( self, DISCONNECT_INDICATION, buf, actual, 
				 final);
		break;
	case CMD_PUT:
		DEBUG( 4, "Got put command\n");
		irobex_do_event( self, PUT_INDICATION, buf, actual, final);
		break;
	case CMD_GET:
		printf( "Got get command\n");
		break;
	case CMD_COMMAND:
		printf( "Got command command\n");
		break;
	case CMD_SETPATH:
		printf( "Got setpath command\n");
		break;
	case CMD_ABORT:
		printf( "Got abort command\n");
		break;
	default:
		/*
		 *  Must be a response (or some garbage :-)
		 */
		irobex_do_event( self, RESPONSE_INDICATION, buf, actual, 
				 final);
	}
	return actual;
}

/*
 * Function irobex_get_file (self)
 *
 *    
 *
 */
void irobex_get_file( IROBEX_HANDLE *self)
{
	while ( irobex_input( self) > 0);
}

/*
 * Function irobex_put_file (self, filename)
 *
 *    Read file and send it to remote device as an IrOBEX object
 *
 */
void irobex_put_file( IROBEX_HANDLE *self, char *filename)
{
	IROBEX_OBJECT *object;
	int fd;
	int file_size;
	int actual;
	struct stat stats;
	char *index = NULL;

	DEBUG( 4, "irobex_put_file()\n");

	fd = open( filename, O_RDONLY, 0);
	if ( fd == -1) {
		perror( "Unable to open file:");
		exit( -1);
	}
	self->fd_file = fd;
	
	/*
	 *  Need to know the file length
	 */
	fstat( fd, &stats);

	DEBUG( 0, "File length=%d\n", (int) stats.st_size);
	file_size = (int) stats.st_size;

	/* Build object */
	object = malloc( sizeof( IROBEX_OBJECT));
	memset( object, 0, sizeof( IROBEX_OBJECT));

	strcpy( object->name, filename);
	strcpy( object->description, filename);
	object->length = file_size;

#ifdef PALM_III
	/* Set Memopad as the default creator ID */
	object->creator_id = MEMO_PAD;
	
	/* Find the . in the filename */
	index = rindex( object->name, '.');
	if ( index) {
		index++;
		if ( strcmp( index, "vcf") == 0) {
			DEBUG( 0, "This is a Address Book file\n");
			object->creator_id = ADDRESS_BOOK;
		} else if ( strcmp( index, "vcs") == 0) {
			DEBUG( 0, "This is a Date Book file\n");
			object->creator_id = DATE_BOOK;
		} else if ( strcmp( index, "txt") == 0) {
			DEBUG( 0, "This is a Memo pad file\n");
			object->creator_id = MEMO_PAD;
		}
	}
#endif
	object->body = malloc( file_size);

	/* Copy file */
	actual = read( self->fd_file, object->body, file_size);
	assert( actual == file_size);
	
	if ( self->object)
		free( self->object);
	self->object = object;

	/*
	 *  Attempt to send the object
	 */
	irobex_put_request( self);
}

/*
 * Function irobex_state_disconnected (self, event, buf, len, final)
 *
 *    IrOBEX layer is disconnected
 *
 */
static int irobex_state_disconnected( IROBEX_HANDLE *self, IROBEX_EVENT event,
				      __u8 *buf, int len, int final)
{
	int ret;

	DEBUG( 4, "irobex_state_disconnected()\n");

	switch( event) {
	case CONNECT_INDICATION:
		irobex_connect_indication( self, buf, len);
		irobex_connect_response( self);
		
		irobex_next_state( self, STATE_READY);
		break;
	case CONNECT_REQUEST:
		ret = irobex_connect_request( self);
		if (ret == -1)
			irobex_next_state( self, STATE_DISCONNECTED);
		else
			irobex_next_state( self, STATE_CONNECTING);
		break;
	case DISCONNECT_INDICATION:
		/* Done that, been there! */
		break;
	case RESPONSE_INDICATION:
		ret = irobex_response_indication( self, buf, len);
		if ( ret == RSP_SUCCESS) {
			/* Should be safe to disconnect now */
			ioctl( self->fd_dev, IROBEX_IOCSDISCONNECT, 
			       &self->daddr);
		}
		irobex_next_state( self, STATE_DISCONNECTED);
		break;
	default:
		DEBUG( 0, "irobex_state_disconnected: Unknown event %s\n",
		       irobex_event[ event]);
		break;
	}
	return 0;
}

/*
 * Function irobex_state_connecting (self, event, buf, len, final)
 *
 *    IrOBEX layer has issued a connect request to the peer device, and is 
 *    awaiting response.
 */
static int irobex_state_connecting( IROBEX_HANDLE *self, IROBEX_EVENT event,
				    __u8 *buf, int len, int final)
{
	DEBUG( 4, "irobex_state_connecting()\n");

	switch( event) {
	case RESPONSE_INDICATION:
		irobex_response_indication( self, buf, len);
		
		/*
		 *  Since the response format for connect responses is quite
		 *  similar to connect requests, we just patch the buffer so
		 *  it looks like a connect indication and let that function
		 *  parse it.
		 */
		buf[0] = 0x00;
		irobex_connect_indication( self, buf, len);

		irobex_next_state( self, STATE_READY);
		irobex_put_request( self);

		break;
	default:
		DEBUG( 0, "irobex_state_connecting: Unknown event %s\n",
		       irobex_event[ event]);
		break;
	}
	return 0;
}

/*
 * Function irobex_state_connected (self, event, buf, len, final)
 *
 *    IrOBEX layer is connected and ready 
 *
 */
static int irobex_state_ready( IROBEX_HANDLE *self, IROBEX_EVENT event,
			       __u8 *buf, int len, int final)
{
	DEBUG( 4, "irobex_state_ready()\n");
	
	switch( event) {
	case PUT_INDICATION:
		irobex_put_indication( self, buf, len, final);

		
		/* Is this the final frame? */
		if ( final) {
			irobex_put_response( self, RSP_SUCCESS, TRUE);

			irobex_next_state( self, STATE_READY);
		} else {
			/* Make progress */
			irobex_put_response( self, RSP_CONTINUE, TRUE);

			irobex_next_state( self, STATE_PUT);
		}
		break;
	case PUT_REQUEST:
		DEBUG( 0, "irobex_state_ready: not implemented\n");
		break;
	case DISCONNECT_INDICATION:
		irobex_disconnect_indication( self, buf, len, final);
	
		irobex_next_state( self, STATE_DISCONNECTED);

		irobex_put_response( self, RSP_SUCCESS, TRUE);
		break;
	case RESPONSE_INDICATION:
		/* TODO: check which response it is */
		
		irobex_next_state( self, STATE_PUT);
		irobex_put_request( self);
		
		break;
	case CONNECT_INDICATION: 
		/* 
		 *  Have we missed the disconnect? No bid deal, just continue
		 *  as normal.
		 */
		irobex_connect_indication( self, buf, len);
		irobex_connect_response( self);
		
		irobex_next_state( self, STATE_READY);
		break;
	default:
		DEBUG( 0, "irobex_state_ready: Unknown event %s\n",
		       irobex_event[ event]);
		break;
	}
	return 0;
}

/*
 * Function irobex_state_put (self, event, buf, len, final)
 *
 *    This is the state which takes care of the put command. Both put 
 *    requests and indications are processed here
 *
 */
static int irobex_state_put( IROBEX_HANDLE *self, IROBEX_EVENT event,
			     __u8 *buf, int len, int final)
{
	DEBUG( 4, "irobex_state_put()\n");
	
	switch( event) {
	case PUT_INDICATION:
		irobex_put_indication( self, buf, len, final);

		/* Is this the final frame */
		if ( final) {
			irobex_put_response( self, RSP_SUCCESS, TRUE);
			irobex_next_state( self, STATE_READY);
			if ( self->callback)
				self->callback( self->object);

			/* 
			 *  Application has now the responsibility for the
			 *  object 
			 */
			self->object = NULL;
		} else {
			/* Make progress */
			irobex_put_response( self, RSP_CONTINUE, TRUE);
			irobex_next_state( self, STATE_PUT);
		}
		break;
	case PUT_REQUEST:
		DEBUG( 0, "irobex_state_put: Not implemented\n");
		break;
	case RESPONSE_INDICATION:
		irobex_response_indication( self, buf, len);
		/* FIXME: Check return value */

		if ( irobex_put_request( self)) {
			irobex_next_state( self, STATE_DISCONNECTED);
			irobex_disconnect_request( self);
		} else
			irobex_next_state( self, STATE_PUT);
		break;
	case DISCONNECT_REQUEST:
		irobex_next_state( self, STATE_DISCONNECTED);
		irobex_disconnect_request( self);
	default:
		DEBUG( 0, "irobex_state_put: Unknown event %s\n",
		       irobex_event[ event]);
		break;
	}
	return 0;
}

/*
 * Function irobex_state_get (self, event, buf, len, final)
 *
 *    
 *
 */
static int irobex_state_get( IROBEX_HANDLE *self, IROBEX_EVENT event,
			     __u8 *buf, int len, int final)
{
	DEBUG( 4, "irobex_state_get()\n");

	switch( event) {
	default:
		DEBUG( 0, "irobex_state_get: Unknown event %s\n",
		       irobex_event[ event]);
		break;
	}
	return 0;
}
